(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translcore.ml,v 1.57 1999/02/24 15:21:44 xleroy Exp $ *)

let generic_instance_debug = try ignore (Sys.getenv "DEBUG_GENERICINSTANCE"); true with _ -> false

(* Translation from typed abstract syntax to lambda terms,
   for the core language *)

open Misc
open Asttypes
open Primitive
open Path
open Types
open Typedtree
open Typeopt
open Lambda

type error =
    Illegal_letrec_pat
  | Illegal_letrec_expr
  | Free_super_var
  | Free_type_variable of type_expr
  | Unknown_generic_primitive of string

exception Error of Location.t * error

(* Forward declaration -- to be filled in by Translmod.transl_module *)
let transl_module =
  ref((fun cc rootpath lambda modl -> assert false) :
      module_coercion -> Path.t option -> lambda -> module_expr -> lambda)

(* Translation of primitives *)

let comparisons_table = create_hashtable 11 [
  "%equal",
      (Pccall{prim_name = "equal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Ceq,
       Pfloatcomp Ceq,
       Pccall{prim_name = "string_equal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal});
  "%notequal",
      (Pccall{prim_name = "notequal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Cneq,
       Pfloatcomp Cneq,
       Pccall{prim_name = "string_notequal"; prim_arity = 2;
              prim_alloc = false; prim_native_name = ""; 
              prim_native_float = false;
	      prim_generic = Normal});
  "%lessthan",
      (Pccall{prim_name = "lessthan"; prim_arity = 2; prim_alloc = false; 
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Clt,
       Pfloatcomp Clt,
       Pccall{prim_name = "lessthan"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal});
  "%greaterthan",
      (Pccall{prim_name = "greaterthan"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Cgt,
       Pfloatcomp Cgt,
       Pccall{prim_name = "greaterthan"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal});
  "%lessequal",
      (Pccall{prim_name = "lessequal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Cle,
       Pfloatcomp Cle,
       Pccall{prim_name = "lessequal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal});
  "%greaterequal",
      (Pccall{prim_name = "greaterequal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal},
       Pintcomp Cge,
       Pfloatcomp Cge,
       Pccall{prim_name = "greaterequal"; prim_arity = 2; prim_alloc = false;
              prim_native_name = ""; prim_native_float = false;
	      prim_generic = Normal})
]

let primitives_table = create_hashtable 31 [
  "%identity", Pidentity;
  "%ignore", Pignore;
  "%field0", Pfield 0;
  "%field1", Pfield 1;
  "%setfield0", Psetfield(0, true);
  "%makeblock", Pmakeblock(0, Immutable);
  "%makemutable", Pmakeblock(0, Mutable);
  "%raise", Praise;
  "%sequand", Psequand;
  "%sequor", Psequor;
  "%boolnot", Pnot;
  "%negint", Pnegint;
  "%succint", Poffsetint 1;
  "%predint", Poffsetint(-1);
  "%addint", Paddint;
  "%subint", Psubint;
  "%mulint", Pmulint;
  "%divint", Pdivint;
  "%modint", Pmodint;
  "%andint", Pandint;
  "%orint", Porint;
  "%xorint", Pxorint;
  "%lslint", Plslint;
  "%lsrint", Plsrint;
  "%asrint", Pasrint;
  "%eq", Pintcomp Ceq;
  "%noteq", Pintcomp Cneq;
  "%ltint", Pintcomp Clt;
  "%leint", Pintcomp Cle;
  "%gtint", Pintcomp Cgt;
  "%geint", Pintcomp Cge;
  "%incr", Poffsetref(1);
  "%decr", Poffsetref(-1);
  "%intoffloat", Pintoffloat;
  "%floatofint", Pfloatofint;
  "%negfloat", Pnegfloat;
  "%absfloat", Pabsfloat;
  "%addfloat", Paddfloat;
  "%subfloat", Psubfloat;
  "%mulfloat", Pmulfloat;
  "%divfloat", Pdivfloat;
  "%eqfloat", Pfloatcomp Ceq;
  "%noteqfloat", Pfloatcomp Cneq;
  "%ltfloat", Pfloatcomp Clt;
  "%lefloat", Pfloatcomp Cle;
  "%gtfloat", Pfloatcomp Cgt;
  "%gefloat", Pfloatcomp Cge;
  "%string_length", Pstringlength;
  "%string_safe_get", Pstringrefs;
  "%string_safe_set", Pstringsets;
  "%string_unsafe_get", Pstringrefu;
  "%string_unsafe_set", Pstringsetu;
  "%array_length", Parraylength Pgenarray;
  "%array_safe_get", Parrayrefs Pgenarray;
  "%array_safe_set", Parraysets Pgenarray;
  "%array_unsafe_get", Parrayrefu Pgenarray;
  "%array_unsafe_set", Parraysetu Pgenarray;
  "%obj_size", Parraylength Pgenarray;
  "%obj_field", Parrayrefu Pgenarray;
  "%obj_set_field", Parraysetu Pgenarray
]

let prim_makearray =
  { prim_name = "make_vect"; prim_arity = 2; prim_alloc = true;
    prim_native_name = ""; prim_native_float = false;
    prim_generic = Normal }

let prim_obj_dup =
  { prim_name = "obj_dup"; prim_arity = 1; prim_alloc = true;
    prim_native_name = ""; prim_native_float = false;
    prim_generic = Normal }

let transl_prim prim args =
  try
    let (gencomp, intcomp, floatcomp, stringcomp) =
      Hashtbl.find comparisons_table prim.prim_name in
    begin match args with
      [arg1; {exp_desc = Texp_construct({cstr_tag = Cstr_constant _}, _)}] ->
        intcomp
    | [{exp_desc = Texp_construct({cstr_tag = Cstr_constant _}, _)}; arg2] ->
        intcomp
    | [arg1; arg2] when has_base_type arg1 Predef.type_int
                     or has_base_type arg1 Predef.type_char ->
        intcomp
    | [arg1; arg2] when has_base_type arg1 Predef.type_float ->
        floatcomp
    | [arg1; arg2] when has_base_type arg1 Predef.type_string ->
        stringcomp
    | _ ->
        gencomp
    end
  with Not_found ->
  try
    let p = Hashtbl.find primitives_table prim.prim_name in
    (* Try strength reduction based on the type of the argument *)
    begin match (p, args) with
        (Psetfield(n, _), [arg1; arg2]) -> Psetfield(n, maybe_pointer arg2)
      | (Parraylength Pgenarray, [arg])   -> Parraylength(array_kind arg)
      | (Parrayrefu Pgenarray, arg1 :: _) -> Parrayrefu(array_kind arg1)
      | (Parraysetu Pgenarray, arg1 :: _) -> Parraysetu(array_kind arg1)
      | (Parrayrefs Pgenarray, arg1 :: _) -> Parrayrefs(array_kind arg1)
      | (Parraysets Pgenarray, arg1 :: _) -> Parraysets(array_kind arg1)
      | _ -> p
    end
  with Not_found ->
    Pccall prim

(* Eta-expand a primitive without knowing the types of its arguments *)

let transl_primitive p =
  let prim =
    try
      let (gencomp, intcomp, floatcomp, stringcomp) =
        Hashtbl.find comparisons_table p.prim_name in
      gencomp
    with Not_found ->
    try
      Hashtbl.find primitives_table p.prim_name
    with Not_found ->
      Pccall p in
  let rec make_params n =
    if n <= 0 then [] else Ident.create "prim" :: make_params (n-1) in
  let params = make_params p.prim_arity in
  Lfunction(Curried, params, Lprim(prim, List.map (fun id -> Lvar id) params))

(* To check the well-formedness of r.h.s. of "let rec" definitions *)

let check_recursive_lambda idlist lam =
  let rec check_top idlist = function
      Lfunction(kind, params, body) as funct -> true
    | Lprim(Pmakeblock(tag, mut), args) ->
        List.for_all (check idlist) args
    | Lprim(Pmakearray(Paddrarray|Pintarray), args) ->
        List.for_all (check idlist) args
    | Llet(str, id, arg, body) ->
        check idlist arg && check_top (add_let id arg idlist) body
    | Lletrec(bindings, body) ->
        let idlist' = add_letrec bindings idlist in
        List.for_all (fun (id, arg) -> check idlist' arg) bindings &&
        check_top idlist' body
    | Levent (lam, _) -> check_top idlist lam
    | _ -> false

  and check idlist = function
      Lvar _ -> true
    | Lconst cst -> true
    | Lfunction(kind, params, body) -> true
    | Llet(str, id, arg, body) ->
        check idlist arg && check (add_let id arg idlist) body
    | Lletrec(bindings, body) ->
        let idlist' = add_letrec bindings idlist in
        List.for_all (fun (id, arg) -> check idlist' arg) bindings &&
        check idlist' body
    | Lprim(Pmakeblock(tag, mut), args) ->
        List.for_all (check idlist) args
    | Lprim(Pmakearray(Paddrarray|Pintarray), args) ->
        List.for_all (check idlist) args
    | Levent (lam, _) -> check idlist lam
    | lam ->
        let fv = free_variables lam in
        List.for_all (fun id -> not(IdentSet.mem id fv)) idlist

  and add_let id arg idlist =
    match arg with
      Lvar id' -> if List.mem id' idlist then id :: idlist else idlist
    | Llet(_, _, _, body) -> add_let id body idlist
    | Lletrec(_, body) -> add_let id body idlist
    | _ -> idlist

  and add_letrec bindings idlist =
    List.fold_right (fun (id, arg) idl -> add_let id arg idl)
                    bindings idlist

  in check_top idlist lam

(* To propagate structured constants *)

exception Not_constant

let extract_constant = function
    Lconst sc -> sc
  | _ -> raise Not_constant

let extract_float = function
    Const_base(Const_float f) -> f
  | _ -> fatal_error "Translcore.extract_float"

(* To find reasonable names for let-bound and lambda-bound idents *)

let rec name_pattern default = function
    [] -> Ident.create default
  | (p, e) :: rem ->
      match p.pat_desc with
        Tpat_var (id,_) -> id
      | Tpat_alias(p, id) -> id
      | _ -> name_pattern default rem

(* Insertion of debugging events *)

let event_before exp lam =
  if !Clflags.debug && lam <> Lstaticfail
  then Levent(lam, {lev_loc = exp.exp_loc.Location.loc_start;
                    lev_kind = Lev_before;
                    lev_repr = None;
                    lev_env = Env.summary exp.exp_env})
  else lam

let event_after exp lam =
  if !Clflags.debug
  then Levent(lam, {lev_loc = exp.exp_loc.Location.loc_end;
                    lev_kind = Lev_after exp.exp_type;
                    lev_repr = None;
                    lev_env = Env.summary exp.exp_env})
  else lam

let event_function exp lam =
  if !Clflags.debug then
    let repr = Some (ref 0) in
    let (info, body) = lam repr in
    (info,
     Levent(body, {lev_loc = exp.exp_loc.Location.loc_start;
                   lev_kind = Lev_function;
                   lev_repr = repr;
                   lev_env = Env.summary exp.exp_env}))
  else
    lam None

(* Translation of expressions *)

let stdlib_prim mdl name = 
  try
    transl_path (fst (Env.lookup_value 
			(Longident.Ldot (Longident.Lident mdl, name))
			Env.empty))
  with Not_found ->
    fatal_error ("Primitive "^ mdl ^"." ^ name ^ " not found.") 

let get_ginfo = function
  | Tpat_any ginfo -> ginfo
  | Tpat_var (_,ginfo) -> ginfo
  | _ -> raise (Dtype.Generic_system_error "Translcore.get_ginfo")
;;

let transl_generic_abstraction patdesc typ comp_func arg =
  let is_generic = Dtype.generalized_dynamic_variables typ <> [] in
  if is_generic then begin
    let ginfo = get_ginfo patdesc in
    let flow_id = Ident.create "flow" in
    ginfo.ginfo_flow <- Some flow_id;
    Lfunction (Curried, [ flow_id ], comp_func arg)
  end else comp_func arg
;;

let transl_instance env invoke path instinfo =
  match instinfo with
  | Inst_normal -> transl_path path
  | Inst_full_generic gflow when not invoke ->
      (* let defined, derived generics *)
      Lapply( transl_path path, [Translflow.transl_flow env gflow] )
  | Inst_full_generic gflow when invoke ->
      (* invoke v gflow *)
      Lapply( stdlib_prim "Generic" "invoke", 
	      [ transl_path path; Translflow.transl_flow env gflow] )
  | Inst_poly_generic { poly_parent= ginfo; poly_pos= pos; poly_call= call } ->
      match ginfo.ginfo_flow with
      |	Some id ->
	  let use_self = 
	    match ginfo.ginfo_layout with
	    | None -> false
	    | Some layout -> List.mem_assoc path layout
	  in
	  begin match use_self, invoke with
	  | false, false ->
             (* let defined, derived generics *)
	      Lapply( transl_path path, [Lprim(Pfield (pos+1), [ Lvar id ]) ])
	  | false, true ->
             (* invoke v gflow *)
	      Lapply( stdlib_prim "Generic" "invoke", 
		      [ transl_path path; Lprim(Pfield (pos+1), [ Lvar id ]) ])
	  | true, true ->
	      let Some self_id = ginfo.ginfo_self in
	      Lapply( transl_path (Pident self_id), 
		                    [ Lprim(Pfield (pos+1), [ Lvar id ])])
	  | _ -> assert false
	  end
      |	None -> 
	  raise (Dtype.Generic_system_error "transl_generic_instance")
;;

let transl_generic_primitive e p instinfo =
  let comp = 
    match p.prim_name with
    | "%dynamic" ->
      	stdlib_prim "Dynamic" "dynamic_comp"
    | "%coerce" ->
	Lapply( stdlib_prim "Dynamic" "coerce_comp", [
	       Lconst (Const_block(0, [
		   Const_base(Const_string !Location.input_name);
		   Const_base(Const_int e.exp_loc.Location.loc_start);
		   Const_base(Const_int e.exp_loc.Location.loc_end) ]))])
    | "%export_value" ->
      	stdlib_prim "Extern" "export_value_comp"
    | "%import_value" ->
	Lapply( stdlib_prim "Extern" "import_value_comp", [
	       Lconst (Const_block(0, [
		   Const_base(Const_string !Location.input_name);
		   Const_base(Const_int e.exp_loc.Location.loc_start);
		   Const_base(Const_int e.exp_loc.Location.loc_end) ]))])
    | "%export_to_string" ->
      	stdlib_prim "Extern" "export_to_string_comp"
    | "%import_from_string" ->
	Lapply( stdlib_prim "Extern" "import_from_string_comp", [
	       Lconst (Const_block(0, [
		   Const_base(Const_string !Location.input_name);
		   Const_base(Const_int e.exp_loc.Location.loc_start);
		   Const_base(Const_int e.exp_loc.Location.loc_end) ]))])
    | _ -> raise (Error (e.exp_loc, Unknown_generic_primitive p.prim_name))
  in 
  match instinfo with
  | Inst_full_generic gflow ->
      (* let defined, derived generics *)
      Lapply( comp, [Translflow.transl_flow e.exp_env gflow] )
  | Inst_poly_generic { poly_parent= ginfo; poly_pos= pos; poly_call= call } ->
      match ginfo.ginfo_flow with
      |	Some id ->
          (* let defined, derived generics *)
	  Lapply( comp, [Lprim(Pfield (pos+1), [ Lvar id ]) ])
      |	None -> 
	  raise (Dtype.Generic_system_error "transl_generic_primitive")
;;

let rec transl_exp e =
  match e.exp_desc with
  | Texp_ident(path, ({val_kind = Val_prim p} as vd), instinforef) when p.prim_generic <> Normal ->
      (* Generic primitives *)
      transl_generic_primitive e p !instinforef
  | Texp_ident(path, {val_kind = Val_prim p},_) ->
      (* Other primitives *)
      transl_primitive p
  | Texp_ident(path, {val_kind = Val_anc _},_) ->
      raise(Error(e.exp_loc, Free_super_var))
  | Texp_ident(path, ({val_kind = Val_reg | Val_self _} as desc),instinforef) ->
(*
      Format.print_string ("transl_generic_instance: " ^ Path.string_of path ^
		     " : " );
      Printtyp.type_expr e.exp_type;
      Format.print_string " (";
      Printtyp.type_scheme desc.val_type;
      Format.print_string ")";
      Format.print_newline ();
*)
      let invoke = 
	match desc.val_konstraints with
	| Konst_Cases _ -> true
	| _ -> false
      in
      transl_instance e.exp_env invoke path !instinforef
  | Texp_constant cst ->
      Lconst(Const_base cst)
  | Texp_let(rec_flag, pat_expr_list, body) ->
      transl_let rec_flag pat_expr_list (event_before body (transl_exp body))
  | Texp_function pat_expr_list ->
      let ((kind, params), body) =
        event_function e
          (function repr ->
             transl_function e.exp_loc !Clflags.native_code repr pat_expr_list)
      in
      Lfunction(kind, params, body)
  | Texp_apply({exp_desc = Texp_ident(path, {val_kind = Val_prim p},_)}, args)
    when List.length args = p.prim_arity && p.prim_generic = Normal -> 
      (* Non Generic primitives *)
      let prim = transl_prim p args in
      let lam = Lprim(prim, transl_list args) in
      begin match prim with Pccall _ -> event_after e lam | _ -> lam end
  | Texp_apply(funct, args) ->
      let lam =
        match transl_exp funct with
          Lsend(lmet, lobj, largs) ->
            Lsend(lmet, lobj, largs @ transl_list args)
        | Levent(Lsend(lmet, lobj, largs), _) ->
            Lsend(lmet, lobj, largs @ transl_list args)
        | Lapply(lexp, largs) ->
            Lapply(lexp, largs @ transl_list args)
        | lexp ->
            Lapply(lexp, transl_list args) in
      event_after e lam
  | Texp_match({exp_desc = Texp_tuple argl} as arg, pat_expr_list) ->
      Matching.for_multiple_match e.exp_loc
        (transl_list argl) (transl_cases pat_expr_list)
  | Texp_match(arg, pat_expr_list) ->
      Matching.for_function e.exp_loc None
        (transl_exp arg) (transl_cases pat_expr_list)
  | Texp_try(body, pat_expr_list) ->
      let id = name_pattern "exn" pat_expr_list in
      Ltrywith(transl_exp body, id,
               Matching.for_trywith (Lvar id) (transl_cases pat_expr_list))
  | Texp_tuple el ->
      let ll = transl_list el in
      begin try
        Lconst(Const_block(0, List.map extract_constant ll))
      with Not_constant ->
        Lprim(Pmakeblock(0, Immutable), ll)
      end
  | Texp_construct(cstr, args) ->
      let ll = transl_list args in
      begin match cstr.cstr_tag with
        Cstr_constant n ->
          Lconst(Const_pointer n)
      | Cstr_block n ->
          begin try
            Lconst(Const_block(n, List.map extract_constant ll))
          with Not_constant ->
            Lprim(Pmakeblock(n, Immutable), ll)
          end
      | Cstr_exception path ->
          Lprim(Pmakeblock(0, Immutable), transl_path path :: ll)
      end
  | Texp_record ((lbl1, _) :: _ as lbl_expr_list, opt_init_expr) ->
      transl_record lbl1.lbl_all lbl1.lbl_repres lbl_expr_list opt_init_expr
  | Texp_field(arg, lbl) ->
      let access =
        match lbl.lbl_repres with
          Record_regular -> Pfield lbl.lbl_pos
        | Record_float -> Pfloatfield lbl.lbl_pos in
      Lprim(access, [transl_exp arg])
  | Texp_setfield(arg, lbl, newval) ->
      let access =
        match lbl.lbl_repres with
          Record_regular -> Psetfield(lbl.lbl_pos, maybe_pointer newval)
        | Record_float -> Psetfloatfield lbl.lbl_pos in
      Lprim(access, [transl_exp arg; transl_exp newval])
  | Texp_array expr_list ->
      let kind = array_kind e in
      let len = List.length expr_list in
      if len <= Config.max_young_wosize then
        Lprim(Pmakearray kind, transl_list expr_list)
      else begin
        let v = Ident.create "makearray" in
        let rec fill_fields pos = function
          [] ->
            Lvar v
        | arg :: rem ->
            Lsequence(Lprim(Parraysetu kind,
                            [Lvar v;
                             Lconst(Const_base(Const_int pos));
                             transl_exp arg]),
                      fill_fields (pos+1) rem) in
        Llet(Strict, v,
             Lprim(Pccall prim_makearray,
                   [Lconst(Const_base(Const_int len));
                    transl_exp (List.hd expr_list)]),
             fill_fields 1 (List.tl expr_list))
      end
  | Texp_ifthenelse(cond, ifso, Some ifnot) ->
      Lifthenelse(transl_exp cond,
                  event_before ifso (transl_exp ifso),
                  event_before ifnot (transl_exp ifnot))
  | Texp_ifthenelse(cond, ifso, None) ->
      Lifthenelse(transl_exp cond,
                  event_before ifso (transl_exp ifso),
                  lambda_unit)
  | Texp_sequence(expr1, expr2) ->
      Lsequence(transl_exp expr1, event_before expr2 (transl_exp expr2))
  | Texp_while(cond, body) ->
      Lwhile(transl_exp cond, event_before body (transl_exp body))
  | Texp_for(param, low, high, dir, body) ->
      Lfor(param, transl_exp low, transl_exp high, dir,
           event_before body (transl_exp body))
  | Texp_when(cond, body) ->
      event_before cond
        (Lifthenelse(transl_exp cond, event_before body (transl_exp body),
                     Lstaticfail))
  | Texp_send(expr, met) ->
      let met_id =
        match met with
          Tmeth_name nm -> Translobj.meth nm
        | Tmeth_val id  -> id
      in
      event_after e (Lsend(Lvar met_id, transl_exp expr, []))
  | Texp_new (cl, _) ->
      Lapply(Lprim(Pfield 0, [transl_path cl]), [lambda_unit])
  | Texp_instvar(path_self, path) ->
      Lprim(Parrayrefu Paddrarray, [transl_path path_self; transl_path path])
  | Texp_setinstvar(path_self, path, expr) ->
      transl_setinstvar (transl_path path_self) path expr
  | Texp_override(path_self, modifs) ->
      let cpy = Ident.create "copy" in
      Llet(Strict, cpy,
           Lapply(Translobj.oo_prim "copy", [transl_path path_self]),
           List.fold_right
             (fun (path, expr) rem ->
                Lsequence(transl_setinstvar (Lvar cpy) path expr, rem))
             modifs
             (Lvar cpy))
  | Texp_letmodule(id, modl, body) ->
      Llet(Strict, id, !transl_module Tcoerce_none None 
	     (* JPF need fix *)
	     (* We have id, but no current module name here... *)
(Lconst(Const_block (0, [Const_base (Const_string "dummy");
					 Const_base (Const_int 0)])))
	     modl, transl_exp body)
      
(**********************
  | Texp_typeinfo path ->
      Transldynamic.transl_typeinfo path
*************************)

  | Texp_generic(rec_flag, env_after, defs, body) ->
      transl_generic rec_flag env_after defs 
	(event_before body (transl_exp body))

  | Texp_coerce(arg, pat_typ_expr_list) ->
      transl_coerce e arg pat_typ_expr_list

  | _ ->
      fatal_error "Translcore.transl"

and transl_list expr_list =
  List.map transl_exp expr_list

and transl_cases pat_expr_list =
  List.map
    (fun (pat, expr) -> (pat, event_before expr (transl_exp expr)))
    pat_expr_list

and transl_tupled_cases patl_expr_list =
  List.map (fun (patl, expr) -> (patl, transl_exp expr)) patl_expr_list

and transl_function loc untuplify_fn repr pat_expr_list =
  match pat_expr_list with
    [pat, ({exp_desc = Texp_function pl} as exp)] ->
    (* case fun (x,y) -> fun (z,w,..) -> *)
      let param = name_pattern "param" pat_expr_list in
      let ((_, params), body) = transl_function exp.exp_loc false repr pl in
      ((Curried, param :: params),
       Matching.for_function loc None (Lvar param) [pat, body])
  | ({pat_desc = Tpat_tuple pl}, _) :: _ when untuplify_fn ->
    (* case function (x,y) -> ... | ... *)
      begin try
        let size = List.length pl in
        let pats_expr_list =
          List.map
            (fun (pat, expr) -> (Matching.flatten_pattern size pat, expr))
            pat_expr_list in
        let params = List.map (fun p -> Ident.create "param") pl in
        ((Tupled, params),
         Matching.for_tupled_function loc params
                                      (transl_tupled_cases pats_expr_list))
      with Matching.Cannot_flatten ->
        let param = name_pattern "param" pat_expr_list in
        ((Curried, [param]),
         Matching.for_function loc repr (Lvar param)
           (transl_cases pat_expr_list))
      end
  | _ ->
      let param = name_pattern "param" pat_expr_list in
      ((Curried, [param]),
       Matching.for_function loc repr (Lvar param)
         (transl_cases pat_expr_list))

and transl_let rec_flag pat_expr_list body =
  match rec_flag with
  | Nonrecursive ->
      let rec transl = function
        [] ->
          body
      | (pat, expr) :: rem ->
	  (* check the types in pat and put identifiers *)
(*
	  let gvars = put_generic_idents pat.pat_type in
	  let body = 
	    match gvars with
	      [] -> transl_exp expr
	    | _ -> Lfunction (Curried, List.map fst gvars, transl_exp expr)
	  in
*)
	  let body = transl_generic_abstraction 
	      pat.pat_desc pat.pat_type transl_exp expr in
          let lam = Matching.for_let pat.pat_loc body pat (transl rem) in
	  (* List.iter (fun (_,x) -> x := Generalized None) gendvars; *)
	  lam
      in transl pat_expr_list
  | Recursive ->
      let idlist =
        List.map
          (fun (pat, expr) -> 
            match pat.pat_desc with
              Tpat_var (id,_) -> id
		  (* possible error Tpat_any ?? *)
            | _ -> raise(Error(pat.pat_loc, Illegal_letrec_pat)))
        pat_expr_list in
      let transl_case (pat, expr) id =
	let lam = transl_generic_abstraction pat.pat_desc
	    pat.pat_type transl_exp expr in
        if not (check_recursive_lambda idlist lam) then
          raise(Error(expr.exp_loc, Illegal_letrec_expr));
	(id, lam)
(*
	(* JPF: we must check id's type before translate expr *)
	let gvars = put_generic_idents pat.pat_type in
	let lam =
	  match gvars with
	    [] -> transl_exp expr
	  | _ -> Lfunction (Curried, List.map fst gvars, transl_exp expr)
	in
	List.iter (fun (_,x) -> x := Generalized None) gvars;
        if not (check_recursive_lambda idlist lam) then
          raise(Error(expr.exp_loc, Illegal_letrec_expr));
        (id, lam) 
*)
      in
      Lletrec(List.map2 transl_case pat_expr_list idlist, body)

and transl_setinstvar self var expr =
  Lprim(Parraysetu (if maybe_pointer expr then Paddrarray else Pintarray),
                    [self; transl_path var; transl_exp expr])

and transl_record all_labels repres lbl_expr_list opt_init_expr =
  (* Determine if there are "enough" new fields *)
  if 3 + 2 * List.length lbl_expr_list >= Array.length all_labels
  then begin
    (* Allocate new record with given fields (and remaining fields
       taken from init_expr if any *)
    let lv = Array.create (Array.length all_labels) Lstaticfail in
    let init_id = Ident.create "init" in
    begin match opt_init_expr with
      None -> ()
    | Some init_expr ->
        for i = 0 to Array.length all_labels - 1 do
          let access =
            match all_labels.(i).lbl_repres with
              Record_regular -> Pfield i
            | Record_float -> Pfloatfield i in
          lv.(i) <- Lprim(access, [Lvar init_id])
        done
    end;
    List.iter
      (fun (lbl, expr) -> lv.(lbl.lbl_pos) <- transl_exp expr)
      lbl_expr_list;
    let ll = Array.to_list lv in
    let mut =
      if List.exists (fun (lbl, expr) -> lbl.lbl_mut = Mutable) lbl_expr_list
      then Mutable
      else Immutable in
    let lam =
      try
        if mut = Mutable then raise Not_constant;
        let cl = List.map extract_constant ll in
        match repres with
          Record_regular -> Lconst(Const_block(0, cl))
        | Record_float ->
            Lconst(Const_float_array(List.map extract_float cl))
      with Not_constant ->
        match repres with
          Record_regular -> Lprim(Pmakeblock(0, mut), ll)
        | Record_float -> Lprim(Pmakearray Pfloatarray, ll) in
    begin match opt_init_expr with
      None -> lam
    | Some init_expr -> Llet(Strict, init_id, transl_exp init_expr, lam)
    end
  end else begin
    (* Take a shallow copy of the init record, then mutate the fields
       of the copy *)
    let copy_id = Ident.create "newrecord" in
    let rec update_field (lbl, expr) cont =
      let upd =
        match lbl.lbl_repres with
          Record_regular -> Psetfield(lbl.lbl_pos, maybe_pointer expr)
        | Record_float -> Psetfloatfield lbl.lbl_pos in
      Lsequence(Lprim(upd, [Lvar copy_id; transl_exp expr]), cont) in
    begin match opt_init_expr with
      None -> assert false
    | Some init_expr ->
        Llet(Strict, copy_id,
             Lprim(Pccall prim_obj_dup, [transl_exp init_expr]),
             List.fold_right update_field lbl_expr_list (Lvar copy_id))
    end
  end

and transl_generic rec_flag body_env pat_expr_list body =
(* 
  let g1_local = [| fun self flow -> ... self flow |]
  and g2_local = ...
  and gn_local = ...
  in
  let g1 = g1_local + mutrec_locals
  
  let gn = gn_local + mutrec_locals
*)  

  let find_pat_expr_of_ident =
    let table =
      List.map (fun (pat,caseset) ->
	match pat.pat_desc with
	| Tpat_var (id,_) -> id, (pat, caseset)
	| _ -> assert false) pat_expr_list
    in
    fun id -> List.assoc id table
  in

  let id_ginfo_of_pat pat =
    match pat.pat_desc with
    | Tpat_var (id,ginfo) -> id, ginfo
    | _ -> assert false 
  in
    
  (* set ginfos *)
  let set_ginfo (pat,caseset) =
    let pat_id, ginfo = id_ginfo_of_pat pat in
    (* get the pattern id and its compilation information storage *)
    (* search mutually recursively defined values *)
    (* set self and flow ids in the compilation information storage *)
    Typegeneric.set_compilation_layout 
      (let vdesc = Env.find_value (Pident pat_id) body_env in 
       vdesc.val_konstraints) ginfo;
    let self_id = Ident.create "self" in (* can I use many times? *)
    let flow_id = Ident.create "flow" in (* can I use many times? *)
    ginfo.ginfo_flow <- Some flow_id;
    ginfo.ginfo_self <- Some self_id
  in
  List.iter set_ginfo pat_expr_list;

  let compilation_cache = ref ([] : (Path.t * 'a) list ) in

  let transl_generic_binding (pat,caseset) =
    (* get the pattern id and its compilation information storage *)
    let pat_id, ginfo = id_ginfo_of_pat pat in
    let Some layout = ginfo.ginfo_layout in
    (* we create an array which contains each compilaiton of type case,
       including those of mutually generics. If such compilation is
       already done, we just copy it. *)

    let transl_compiled_binding path pos len =
      let comp_path = transl_path path in
      let rec sub x = 
	if x = len then []
	else Lprim(Pfield (pos+x), [comp_path]) :: sub (x + 1)
      in
      sub 0, len
    in
      
    let transl_outside_compiled_binding path =
      let len = 
	let vdesc = Env.find_value path body_env in
	match vdesc.val_konstraints with
	| Konst_Cases (p, cks) -> List.length cks
	| _ -> assert false
      in
      transl_compiled_binding path 0 len
    in

    let transl_typecase self_id flow_id cas =
      Lfunction (Curried, [ self_id; flow_id ], transl_exp cas.typecase_method)
    in

    let transl_new_binding pos id =
      let pat, caseset = find_pat_expr_of_ident id in
      let _, ginfo = id_ginfo_of_pat pat in
      let Some flow_id = ginfo.ginfo_flow
      and Some self_id = ginfo.ginfo_self
      in
      let comp = 
         List.flatten (List.map (function 
          | Tgeneric_typecase cas ->
              [transl_typecase self_id flow_id cas]
          | Tgeneric_include (path,vdesc,loc) -> 
              fst (transl_outside_compiled_binding path)
        ) caseset.gen_cases)
      in
      let len = List.length comp in
      (* store it in cache *)
      compilation_cache := 
	(Pident id, (Pident pat_id, pos, len)) :: !compilation_cache;
      comp, len
    in

    let rec transl_binding pos = function
      |	[] -> []
      |	(path,pos')::ps ->
          (* we double check pos *)
          assert (pos = pos');
	  let comp_path, len = 
	    try
	      (* if it is in cache *)
	      let path', pos, len = List.assoc path !compilation_cache in
	      transl_compiled_binding path' pos len
	    with
	    | Not_found ->
	        (* if it is not in cache *)
		match path with
		| Pident id ->
		    begin try
		      (* it is defined here, not compiled yet *)
		      transl_new_binding pos id
		    with
		    | Not_found ->  
			(* it is defined outside *)
			transl_outside_compiled_binding path
		    end
		| _ -> transl_outside_compiled_binding path

	  in
	  comp_path @ transl_binding (pos + len) ps
    in

    pat_id, 
    Lprim (Pmakeblock (0, Immutable), transl_binding 0 layout)
  in
  
  let rec transl = function
    | [] -> body
    | binding :: rem ->
	let id, comp = transl_generic_binding binding in
	Llet( Strict, id, comp, transl rem)
  in
  transl pat_expr_list

and transl_coerce e arg pat_typ_expr_list =
  (* It is really look alike with Texp_match, but its compilation
     cannot be done with match compiler *)
  (* Current implementation have no optimization *)
  let id_dyn = Ident.create "dyn" 
  and id_val = Ident.create "val" 
  and id_rtyp = Ident.create "rtyp"
  in

  let action_compiler loc untuplify_fn repr pat expr =
    match Matching.for_function loc None (Lvar id_val) 
	(transl_cases [pat,expr]) with
      Lcatch (lambda, _) -> lambda
    | lambda -> lambda
  in
  let rec case_compiler = function
    | (vpat,tpat,action) :: xs ->
	Transltype.inside_coerce_typecase := true; 
	let rtpat = 
	  try
	    let result = Transltype.transl_rtype_of_type e.exp_env tpat in
	    Transltype.inside_coerce_typecase := false;
	    result
	  with
	    e -> 
	      Transltype.inside_coerce_typecase := false;
	      raise e
	in
	Lcatch (
	  Lifthenelse (Lapply( stdlib_prim "Dynamic" "is_instance",
			      [ Lvar id_rtyp; rtpat ]),
	               (* like fun vpat -> action, but we exit, instead of
			  raise *)
		       action_compiler e.exp_loc (* incorrect *)
			 !Clflags.native_code None vpat action,
		       Lstaticfail ),
	  case_compiler xs )
    | [] -> Lstaticfail
  in
  Llet(Strict, id_dyn, transl_exp arg,
    Llet(Strict, id_val, Lprim(Pfield 0, [Lvar id_dyn]),
      Llet(Strict, id_rtyp, Lprim(Pfield 1, [Lvar id_dyn]),
        Lcatch (case_compiler pat_typ_expr_list,
                (* JPF: stupid. I don't know how to raise it directly *)
		Lapply( stdlib_prim "Dynamic" "fail",
		       [ Lvar id_rtyp;
			 Lconst (Const_base(Const_string !Location.input_name));
			 Lconst (Const_base(Const_int e.exp_loc.Location.loc_start));
			 Lconst (Const_base(Const_int e.exp_loc.Location.loc_end)) ])))))
;;

let transl_eval flowptr scm expr = (* for Tstr_eval *)
  (* we need generalized type scheme of expr. expr.exp_type is not
       generalized *)
  transl_generic_abstraction (Tpat_any flowptr) scm transl_exp expr
(*
  (* check the types and put identifiers *)
  let gvars = put_generic_idents expr.exp_type in
  let lam =
    match gvars with
      [] -> transl_exp expr
    | _ -> Lfunction (Curried, List.map fst gvars, transl_exp expr)
  in
  List.iter (fun (_,x) -> x := Generalized None) gvars;
  lam
*)

(* Compile an exception definition *)

let transl_exception id path decl =
  let name =
    match path with
      None -> Ident.name id
    | Some p -> Path.name p in
  Lprim(Pmakeblock(0, Immutable), [Lconst(Const_base(Const_string name))])

(* Error report *)

open Format

let report_error = function
    Illegal_letrec_pat ->
      print_string
      "Only variables are allowed as left-hand side of `let rec'"
  | Illegal_letrec_expr ->
      print_string
      "This kind of expression is not allowed as right-hand side of `let rec'"
  | Free_super_var ->
      print_string
        "Ancestor names can only be used to select inherited methods"
  | Free_type_variable typ ->
      print_string "This expression has a type ";
      Printtyp.type_scheme typ;
      print_string " , which contains free variable(s) and cannot be converted into a runtime type"
  | Unknown_generic_primitive name ->
      print_string "Generic polymorphism primitive ";
      print_string name;
      print_string " is not defined."
