open Misc
open Asttypes
open Longident
open Lambda
open Types
open Path
open Typedtree
open Transltools

open Rdigest

let dbg = try ignore (Sys.getenv "DEBUG_TRANSLTYPE"); true with _ -> false

(***************************************************************** exception *)

exception Free_type_variable of type_expr

(************************************* some flags and states for compilation *)

let inside_coerce_typecase = ref false 

(******************************************** type --> rpath rtype converter *)

let rpath_of_predefined_type id =
  RPident (Ident.name id, Ident.stamp id)

let rec rpath_of_path = function
    Pident i -> 
      RPident (Ident.name i, Ident.stamp i)
  | Pdot (p,n,x) -> 
      RPdot (rpath_of_path p, n, x)
  | Papply (p1,p2) -> RPapply (rpath_of_path p1, rpath_of_path p2)

let rec transl_rpath = function
    RPident (i,s) -> 
      Const_block (0, [Const_base (Const_string i);
		       Const_base (Const_int s)])
  | RPdot (p,n,x) ->
      Const_block (1, [transl_rpath p; Const_base (Const_string n); Const_base (Const_int x)])
  | RPapply (p1,p2) -> 
      Const_block (2, [transl_rpath p1; transl_rpath p2])

let transl_rpath_of_ident modname id =
  (* RPdot (modid, Ident.name id, Ident.stamp id) *)
  Lprim( Pmakeblock (1, Immutable), [ 
	modname;
	Lconst (Const_base (Const_string (Ident.name id)));
	Lconst (Const_base (Const_int (Ident.stamp id))) ] )

(*********************************************************** lazy evaluation *)
(*
let make_delayed lam =
  Lprim( Pmakeblock (0, Mutable), [
	Lprim( Pmakeblock (0, Immutable), [
	      Lfunction (Curried, [Ident.create "unit"], lam) ] ) ] )

*)
let make_value lam =
  try
    Lconst(Const_block (0, [extract_constant lam]))
  with
    Not_constant ->
      Lprim( Pmakeblock (0, Immutable), [ lam ] )

(* FORCE EAGER EVALUATION *)
(* let make_delayed = make_value *)

let dummy_rpath_comp = 
  Lconst (Const_block (0, [ Const_base (Const_string "dummy");
			    Const_base (Const_int 0) ]))
let dummy_digest_comp =
  Lconst (Const_block (0, [Const_base (Const_string "dummy")]))

let make_type_info modname id digest =
  Lprim( Pmakeblock (0, Mutable), [ 
	transl_rpath_of_ident modname id; 
	make_value digest])

(***********************************************************/lazy evaluation *)

(***************************************************************** type_info *)

let transl_type_info path =
  try
    match path with
      Pident id -> 
	let digest = List.assoc id Predef.builtin_types in
	Lprim( Pmakeblock (0, Mutable), [
	      Lconst (transl_rpath (rpath_of_predefined_type id));
	      Lconst (Const_block (0, [
				   Const_block (0, [Const_base (Const_string digest)]) ]))])
	  (* Possibility of bug... *)
    | _ -> raise Not_found
  with
    Not_found -> Lambda.transl_path path

let transl_type_name path =
  try
    match path with
      Pident id when List.mem_assoc id Predef.builtin_types -> 
	Lconst (transl_rpath (rpath_of_predefined_type id))
    | _ -> raise Not_found
  with
    Not_found -> 
      Lprim(Pfield 0, [ transl_type_info path ])

let type_name_of_info lam =
  Lprim(Pfield 0, [ lam ])

(*********************************************** create rtype and compile it *)

(* We have a type expression, compile the runtime representation for it *)
(* env is needed if we want to expand it *)
 
let transl_rtype_of_type env t =
  if dbg then begin
    Format.print_string "transl_rtype ";
    Printtyp.type_expr t;
    Format.print_newline ();
  end;
  let varcnt = ref 1 (* must begin with 1 *)
  and vartbl = ref [] in
  
  let rec sub t =
    let new_type_var t = 
      (* RTvar (!varcnt) *)
      let rt = 
	Lconst (Const_block (0, [Const_base (Const_int !varcnt)])) 
      in
      (* cache it *)
      vartbl := (t, rt) :: !vartbl;
      incr varcnt;
      rt
    in

    let any_type_var t = (* for pattern creation *) 
      (* RTvar (-!varcnt) *)
      let rt = 
	Lconst (Const_block (0, [Const_base (Const_int (- !varcnt))]))
      in
      (* cache it *)
      vartbl := (t, rt) :: !vartbl;
      incr varcnt;
      rt
    in

    let t = Ctype.repr t in
    match t.desc with
    | Tdvar (x,y) ->
	begin
	  match !x,!y with
	  | Generalized, _ 
	  | _, Exportable ->
	      (try List.assq t !vartbl with Not_found -> new_type_var t)
	  | _ ->
	      prerr_endline "Runtime type creation for unsafe dvar";
	      assert false
	end
    | Tvar -> (* BUG?: normal tvar cannot appear here... *)
	(try List.assq t !vartbl with Not_found -> new_type_var t)
    | Tany -> (* FOR COERCE, WE DO A DIRTY HACK *)
	any_type_var t
    | Tarrow (f,t) -> 
	(* RTarrow (sub f, sub t) *)
	begin
	  let ll = [sub f; sub t] in
	  try
	    Lconst (Const_block (1, List.map extract_constant ll))
	  with
	    _ -> Lprim(Pmakeblock(1, Immutable), ll)
	end
    | Ttuple tl -> 
	(* RTtuple (List.map sub tl) *)
	begin
	  let ll = [transl_list sub tl] in
	  try
	    Lconst (Const_block (2, List.map extract_constant ll))
	  with
	    _ -> Lprim(Pmakeblock(2, Immutable), ll)
	end
    | Tconstr (p,tl,abbrev) -> 
(******************************************************** expansion is disabled
	(* try to expand *)
	begin
	  try
	    let t' = Ctype.expand_abbrev env p tl abbrev t.level in
	    sub t'
	  with Ctype.Cannot_expand ->
            let ll = [ Lconst (transl_rpath_of_path env p); transl_list sub tl ]
	    in
	    try
	      Lconst (Const_block (3, List.map extract_constant ll))
		with
	      _ -> Lprim(Pmakeblock(3, Immutable), ll)
	end
******************************************************************************)
	begin
(************************************************************************** OLD
	  let tinfo = 
	    let ll = [ Lconst (transl_rpath_of_path env p); transl_type_info p ] in
	    try
	      Lconst (Const_block (0, List.map extract_constant ll))
	    with
	      _ -> Lprim (Pmakeblock (0, Immutable), ll)
	  in
******************************************************************************)
	  let tinfo = transl_type_info p in
	  let ll = [ tinfo; transl_list sub tl ]
	  in
	  try
	    Lconst (Const_block (3, List.map extract_constant ll))
	  with
	    _ -> Lprim(Pmakeblock(3, Immutable), ll)
	end
		
    | Tlink t -> sub t
    | Tobject _ 
    | Tfield _
    | Tnil -> 
	raise (Dtype.Generic_unsupported_error "Actually, I do not know about object type system") 
  in
  sub t

let stdlib_prim mdl name = 
  try
    transl_path (fst (Env.lookup_value 
			(Longident.Ldot (Longident.Lident mdl, name))
			Env.empty))
  with Not_found ->
    fatal_error ("Primitive "^ mdl ^"." ^ name ^ " not found.") 

let transl_rtype translator typ = 
  let rec sub = function
    RTvar x -> 
      Const_block (0, [Const_base (Const_int x)])
  | RTarrow (f,t) ->
      Const_block (1, [sub f; sub t])
  | RTtuple typs ->
      Const_block (2, [transl_const_list sub typs])
  | RTconstr (v, typs) ->
      Const_block (3, [translator v; transl_const_list sub typs])
  in
  sub typ

let type_expr_of_rtype rt =
  let tvars = ref [] in
  let rec path_of_rpath = function
      RPident (name,stamp) -> 
	Pident (Ident.create_with_stamp name stamp)
    | RPdot (RPident ("*toplevel*",0), s, x) ->
	prerr_endline (Printf.sprintf "%s %d" s x);
	Pident (Ident.create_with_stamp s x)
    | RPdot (rp, s, x) ->
	prerr_endline s;
	Pdot (path_of_rpath rp, s, x)
    | RPapply (p1,p2) -> 
	Papply (path_of_rpath p1, path_of_rpath p2)
  in
  let rec sub = function
    | RTvar x -> 
	begin 
	  try List.assoc x !tvars with Not_found ->
	    let tv = Btype.newgenty Tvar in
	    tvars := (x,tv) :: !tvars;
	    tv
	end
    | RTarrow (f,t) ->
	Btype.newgenty (Tarrow (sub f, sub t))
    | RTtuple tls ->
	Btype.newgenty (Ttuple (List.map sub tls))
    | RTconstr ((rp,info), args) ->
	Btype.newgenty (Tconstr (path_of_rpath rp, 
				 List.map sub args, ref Mnil))
  in
  sub rt
;;
