open Misc
open Asttypes
open Types
open Btype
open Ctype

let debug = try ignore (Sys.getenv "DEBUG_DTYPE"); true with _ -> false

exception Generic_unsupported_error of string
exception Generic_system_error of string

                  (**********************************************)
                  (*              Anti-unification              *)
                  (**********************************************)

(* anti_unify receives CLOSED type scheme t1 and t2, and returns their
   least general antiunifier. The levels of the newly created
   type variables have the level !current_level.
   You may have to generalize it later.
*)

let anti_unify env t1 t2 =
  let cache = ref [] in
  let add_and_return t1 t2 t =
    cache := ((t1,t2), t) :: !cache;
    t
  in
  let rec sub t1 t2 = 
    let t1 = repr t1
    and t2 = repr t2 in
    let found = ref None in 
    try
      List.iter (fun ((t1',t2'),t) ->
	if equal env false [t1;t2] [t1';t2'] then begin
	  found := Some t;
	  raise Exit
	end) !cache;
      raise Not_found
    with
      Exit -> 
	begin match !found with
	  Some t -> t
	| _ -> assert false
	end
    | Not_found ->
        match t1.desc, t2.desc with
	  Tarrow (t11, t12), Tarrow (t21, t22) ->
	    newty (Tarrow( sub t11 t21, sub t12 t22 ))
	| Ttuple tl1, Ttuple tl2 when List.length tl1 = List.length tl2 ->
	    newty (Ttuple( List.map2 sub tl1 tl2))
	| Tconstr (p1, tl1, _), Tconstr (p2, tl2, _) when Path.same p1 p2 ->
	    (* abbrev? *)
	    newty (Tconstr (p1, List.map2 sub tl1 tl2, ref Mnil (* ??? *) ))
	| _, _ -> 
	    add_and_return t1 t2 (newdvar ())
  in
  sub t1 t2
;;

(* Anti-unification of set of types *)
let anti_unify_types env types = 
  List.fold_right (fun typ st ->
    anti_unify env typ st) (List.tl types) (List.hd types)
;;

                  (**********************************************)
                  (*  Miscellaneous operations on dynamic types *)
                  (**********************************************)

let rec dynamize ty =
  let ty = repr ty in
  match ty.desc with
  | Tvar -> ty.desc <- Tdvar (ref Not_generalized, ref Not_exportable)
  | Tlink ty -> dynamize ty (* actually unused *) 
  | _ -> iter_type_expr dynamize ty

let _ = Ctype.dynamize_ref := dynamize

(* GENERIC: type variable extraction 
     These functions are bit inefficient.
*)

let type_variables n d t =
  let vars = ref [] in
  let rec sub t = 
    let t = repr t in
    match t.desc with
    | Tdvar _ -> if d && not (List.memq t !vars) then vars := t :: !vars
    | Tvar -> if n && not (List.memq t !vars) then vars := t :: !vars
    | Tany -> ()
    | Tarrow (f,t) -> sub f; sub t
    | Ttuple tl -> List.iter sub tl
    | Tconstr (_,tl,_) -> List.iter sub tl
    | Tlink t -> sub t
    | Tobject _ -> raise (Generic_unsupported_error "Do not think about objects yet")
    | Tfield _ -> raise (Generic_unsupported_error "I hate objects")
    | Tnil -> raise (Generic_unsupported_error "Hey, objects are not supported")
  in
  sub t;
  List.rev !vars

let normal_variables = type_variables true false 
let dynamic_variables = type_variables false true
let type_variables = type_variables true true
let generalized_dynamic_variables typ =
  List.fold_right (fun d st ->
    if d.level = Btype.generic_level then d :: st else st)
      (dynamic_variables typ) []

let equal_scheme env scm1 scm2 =
  let abs_scheme scm =
    List.fold_right (fun v st ->
      if v.level = Btype.generic_level then v :: st else st)
      (type_variables scm) [scm]
  in
  let absscm1 = abs_scheme scm1 
  and absscm2 = abs_scheme scm2
  in
(*
  Format.print_string "equal_scheme ";
  Format.print_newline ();
  List.iter (fun t -> !print_scheme t) absscm1;
  Format.print_newline ();
  List.iter (fun t -> !print_scheme t) absscm2;
  Format.print_newline ();
*)
  Ctype.equal env true absscm1 absscm2
;;

                           (***********************)
                           (* Copy generalization *)
                           (***********************)

(* GENERIC: Read the comment in types.mli Tdvar *)

let generalized_variables = ref []

let inplacable ty = 
  let cache = ref [] in
  let rec sub ty =
    try List.assq ty !cache with Not_found ->
      let v =
	match ty.desc with
	| Tvar -> true
	| Tdvar _ -> false
	| Tarrow (t1, t2) -> 
	    let v1 = sub t1 
	    and v2 = sub t2
	    in v1 && v2
	| Ttuple tl ->
	    let vl = List.map (fun t -> sub t) tl in
	    List.fold_right (fun v st -> v && st) vl true
	| Tconstr (p, tl, abbrev) ->
	    let vl = List.map (fun t -> sub t) tl in
	    List.fold_right (fun v st -> v && st) vl true
	| Tobject (t1, nameref) -> sub t1
	| Tfield (label, kind, t1, t2) -> 
	    let v1 = sub t1
	    and v2 = sub t2
	    in v1 && v2
	| Tnil -> true
	| Tlink t -> sub t
	| Tany -> true
      in
      (* we update the cache *)
      cache := (ty,v) :: !cache;
      begin match ty.desc with
      |	Tconstr (_,_,abbrev) -> ignore (sub_expans !abbrev)
      |	_ -> ()
      end;
      v
  and sub_expans = function
    | Mnil -> true
    | Mcons(p, ty1, ty2, rem) ->
	let v1 = sub ty1
	and v2 = sub ty2
	and v' = sub_expans rem 
	in v1 && v2 && v'
    | Mlink rem -> sub_expans !rem 
  in
  sub ty;
  !cache
;;

let rec generalize_copy ty =
  (* even if the level is already equal to generic_level,
     Tdvar link may still be Not_generalized. *)

  let cache = ref [] in
  let inplacable = inplacable ty in

  let rec inplace ty = 
    if not (List.mem_assq ty !cache) then begin
      cache := (ty,ty)::!cache;
      if ty.level > !current_level then begin
	ty.level <- generic_level;
	begin match ty.desc with
	| Tvar -> 
	    generalized_variables := (ty,ty)::!generalized_variables
	| Tconstr(_,_,abbrev) ->
	    inplace_expans !abbrev
	| _ -> ()
	end;
	iter_type_expr inplace ty
      end
    end
  and inplace_expans = function
    | Mnil -> ()
    | Mcons(_, ty1, ty2, rem) -> 
	inplace ty1; 
	inplace ty2; 
	inplace_expans rem
    | Mlink rem -> inplace_expans !rem
  
  and copy ty =
    if ty.level > !current_level then begin
      let t = newgenvar () in (* make a copy *)
      t.level <- generic_level;
      cache := (ty, t) :: !cache;
      t.desc <-
	 begin match ty.desc with
	 | Tvar -> Tvar
	 | Tdvar (x,di) -> 
	     generalized_variables := (ty,t)::!generalized_variables;
    	     x := Generalized;
    	     Tdvar (x,di)
	 | Tarrow (t1, t2) ->
	     Tarrow( gencopy t1, gencopy t2 )
	 | Ttuple tl ->
	     Ttuple( List.map gencopy tl )
	 | Tconstr (p, tl, abbrev) ->
	     Tconstr( p, List.map gencopy tl, ref (copy_expans !abbrev))
	 | Tobject (t1, nameref) ->
	     Tobject( gencopy t1, nameref );
	 | Tfield (label, kind, t1, t2) ->
	     Tfield( label, kind, gencopy t1, gencopy t2 )
	 | Tnil -> assert false
	 | Tlink t -> Tlink( gencopy t )
	 | Tany -> assert false
	 end;
      t
    end else begin
      cache := (ty,ty) :: !cache;
      ty
    end
  and copy_expans = function
    | Mnil -> Mnil
    | Mcons(p, ty1, ty2, rem) ->  
	Mcons(p, gencopy ty1, gencopy ty2, copy_expans rem)
    | Mlink rem -> Mlink(ref (copy_expans !rem))

  and gencopy ty =
    if (ty.level > !current_level) (* && (ty.level <> generic_level) *)
    then begin
      try 
	List.assq ty !cache
      with Not_found ->
	if try List.assq ty inplacable with Not_found -> false then begin 
	  inplace ty; ty 
	end else begin
	  copy ty
	end
    end else ty

  in
  gencopy ty
;;

let generalize_copy ty =
  generalized_variables := [];
  if debug then begin
    !print_type ty;
    Format.print_string " => ";
  end;
  let ty' = generalize_copy ty in
  if debug then begin
    !print_type ty';
    Format.print_newline ()
  end;
  ty'

(* force_generalize_copy generalizes all the type variables 
   regardless of their levels. Use with CARE. *)
let force_generalize_copy ty =
  let level = !current_level in
  current_level := 0;
  let ty = generalize_copy ty in
  current_level:= level;
  ty

let generalized_variables () =
  let result = !generalized_variables in
  generalized_variables := []; 
  (* need to reset, because sometimes this function can be called
     without preceding generalize_copy (when nonexpansive) *)
  result

(********************************************************************* Tools *)

let untuple t = 
  match t.desc with
  | Ttuple l -> l
  | _ -> assert false

(**************************************************************** Copy types *)
(* unlike Ctype.copy, they create a new variable for every variable 
   (no matter if it is generalized or not) .
*)

(* Reset the dynamic variable links. Must be careful to use. Destructive *) 
(*
let reset_tdvar_links typ =
  let dvars = dynamic_variables typ in
  List.iter (fun dvar ->
    match dvar.desc with
    | Tdvar x -> dvar.desc <- Tdvar (ref Not_generalized)
    | _ -> assert false) dvars
;;
*)

type mode_for_links = Keep | Duplicate | Reset

type copy_cache = { 
    mutable copy: (type_expr * type_expr) list;
    mutable link: (tdvar_info ref * tdvar_info ref) list;
    mode: mode_for_links
  } 

let create_copy_cache mode = { copy= []; link= []; mode= mode }

let copy_type cache typ =

  let rec sub ty =
    let ty = repr ty in
    try List.assq ty cache.copy with
    | Not_found ->
    	let t = newgenvar () in
    	t.level <- ty.level;
	cache.copy <- (ty,t) :: cache.copy;
	t.desc <- 
	   begin match ty.desc with
	   | Tvar -> Tvar
	   | Tdvar (x,di) -> 
	       let x' =
		 try List.assq x cache.link with
		 | Not_found ->
		     let x' =
		       match cache.mode with
		       | Keep -> x
		       | Duplicate -> ref !x
		       | Reset -> ref Not_generalized
		     in
		     cache.link <- (x,x') :: cache.link;
		     x'
	       in
	       Tdvar (x',di)
	   | Tarrow (t1, t2) -> Tarrow( sub t1, sub t2 )
	   | Ttuple tl -> Ttuple (List.map sub tl)
	   | Tconstr (p, tl, abbrev) ->
	       Tconstr (p, List.map sub tl, ref (copy_expans !abbrev))
	   | Tobject (t1, {contents = name}) ->
               let name' =
		 match name with
		 | None -> None
		 | Some (p, tl) -> Some (p, List.map sub tl)
               in
               Tobject (sub t1, ref name')
	   | Tfield (label, kind, t1, t2) ->
               begin match field_kind_repr kind with
		 Fpresent -> Tfield (label, Fpresent, sub t1, sub t2)
               | _ -> Tlink (sub t2)
               end
	   | Tnil ->
               Tnil
	   | Tlink t -> (* Actually unused *)
               Tlink (sub t)
	   | Tany -> Tany
	   end;
	t
  and copy_expans = function
    | Mnil ->  Mnil
    | Mcons(p, ty, ty', rem) ->  
      	Mcons(p, sub ty, sub ty', copy_expans rem)
    | Mlink rem         ->  
      	Mlink(ref (copy_expans !rem))
  in

  sub typ
;;  

let copy_types mode_for_links typs =
  let cache = create_copy_cache mode_for_links in
  List.map (copy_type cache) typs
;;

   			     (***************)
   			     (*  Type sizes *)
   			     (***************)

type size_dvar = {
    info: tdvar_info ref;
    mutable num: int
  } 

type size = {
    mutable tdvars: size_dvar list;
    mutable const: int
  }

let string_of_size s =
  List.fold_right (fun tdvar st ->
    st ^ string_of_int tdvar.num ^ 
    "x" ^ string_of_int (Obj.magic tdvar.info mod 100) ^ " + " ) s.tdvars ""
    ^
  string_of_int s.const
;;

let create_size () = { tdvars= []; const= 0 };;
  
let rec size_assq inforef = function
  | [] -> raise Not_found
  | x :: xs -> if x.info == inforef then x else size_assq inforef xs
;;

let add_tdvar s inforef v =
  try
    let sdvar = size_assq inforef s.tdvars in
    sdvar.num <- sdvar.num + v
  with
  | Not_found ->
      s.tdvars <- {info= inforef; num= v} :: s.tdvars
;;

let add_const s v = s.const <- s.const + v;;
  
let rec size_type s t = 
  match t.desc with
  | Tvar -> add_const s 1
  | Tdvar (inforef,_) -> add_tdvar s inforef 1
  | Tarrow (t1,t2) -> size_type s t1; size_type s t2; add_const s 1
  | Ttuple tls -> 
      List.iter (fun t -> size_type s t) tls; 
      add_const s 1 (* for constructor * *)
  | Tconstr (p, tls, _) ->
      List.iter (fun t -> size_type s t) tls; 
      add_const s 1 (* for constructor p *)
  | Tobject _ | Tfield _ | Tnil -> assert false
  | Tlink t -> size_type s t
  | Tany -> assert false
;;

let size t =
  let s = create_size () in
  size_type s t;
  s
;;

(* |t1 <: t2 K| *) 
let size_of_instance_predicate env t scm =
(*
  Format.print_string "|";
  !print_scheme t;
  Format.print_string " <: ";
  !print_scheme scm;
  Format.print_string "| = ";
*)
  let [t';scm'] = copy_types Keep [t;scm] in
  let scm' = instance scm' in
  let dvars = dynamic_variables scm' in
  (* POSSIBLE BUG: non generalized dynamic variables are also counted!!! *)
  try
    Ctype.unify env t' scm';
    let s = create_size () in
    List.iter (size_type s) dvars;
(*
    Format.print_string (string_of_size s);
    Format.print_newline ();
*)
    s
  with
    _ -> raise (Generic_system_error "size_of_instance_predicate")
;;
  
let leq s1 s2 =
  (* s1 is smaller than or equal to s2 ? *)
(*
  prerr_endline (Printf.sprintf "%s <= %s" 
		   (string_of_size s1)
		   (string_of_size s2));
*)
  let tdvars_s1_s2 =
    List.fold_right (fun tdvar st ->
      if List.memq tdvar.info st then st else tdvar.info :: st)
      (s1.tdvars @ s2.tdvars) []
  in
  let sum = ref 0 in
  try
    List.iter (fun tdvar ->
      let sd1 = try (size_assq tdvar s1.tdvars).num with Not_found -> 0 in
      let sd2 = try (size_assq tdvar s2.tdvars).num with Not_found -> 0 in
      if sd2 - sd1 < 0 then raise Exit;
      sum := !sum + sd2 - sd1) tdvars_s1_s2;
    !sum + s2.const - s1.const >= 0
  with
  | Exit | Not_found -> false

