(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: primitive.ml,v 1.4 1997/05/13 14:06:56 xleroy Exp $ *)

(* Description of primitive functions *)

open Misc
open Format

type generic_flag =
    Normal
  | Generic
  | Generic_dynamic
  | Generic_coerce

type description =
  { prim_name: string;         (* Name of primitive  or C function *)
    prim_arity: int;           (* Number of arguments *)
    prim_alloc: bool;          (* Does it allocates or raise? *)
    prim_native_name: string;  (* Name of C function for the nat. code gen. *)
    prim_native_float: bool;   (* Does the above operate on unboxed floats? *)
    prim_generic: generic_flag }       (* Is it a generic primitive ? *)

let parse_declaration arity decl =
  match decl with
    name :: "noalloc" :: name2 :: "float" :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = false;
       prim_native_name = name2; prim_native_float = true;
       prim_generic = Normal}
  | name :: "noalloc" :: name2 :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = false;
       prim_native_name = name2; prim_native_float = false;
       prim_generic = Normal}
  | name :: name2 :: "float" :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = true;
       prim_native_name = name2; prim_native_float = true;
       prim_generic = Normal}
  | name :: "noalloc" :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = false;
       prim_native_name = ""; prim_native_float = false;
       prim_generic = Normal}
  | name :: "generic" :: "dynamic" :: _ ->
      {prim_name = name; prim_arity = 0; prim_alloc = false;
       prim_native_name = ""; prim_native_float = false;
       prim_generic = Generic_dynamic}
  | name :: "generic" :: "coerce" :: _ ->
      {prim_name = name; prim_arity = 0; prim_alloc = false;
       prim_native_name = ""; prim_native_float = false;
       prim_generic = Generic_coerce}
  | name :: "generic" :: _ ->
      {prim_name = name; prim_arity = 0; prim_alloc = false;
       prim_native_name = ""; prim_native_float = false;
       prim_generic = Generic}
  | name :: name2 :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = true;
       prim_native_name = name2; prim_native_float = false;
       prim_generic = Normal}
  | name :: _ ->
      {prim_name = name; prim_arity = arity; prim_alloc = true;
       prim_native_name = ""; prim_native_float = false;
       prim_generic = Normal}
  | [] ->
      fatal_error "Primitive.parse_declaration"

let print_quoted s = print_char '"'; print_string s; print_char '"'

let print_description p =
  print_quoted p.prim_name;
  if not p.prim_alloc then
    (print_space(); print_quoted "noalloc");
  if p.prim_native_name <> "" then
    (print_space(); print_quoted p.prim_native_name);
  if p.prim_native_float then
    (print_space(); print_quoted "float");
  match p.prim_generic with
    Generic -> print_space(); print_quoted "generic"
  | _ -> ()

