(* digest calculation at type inference *)

open Types
open Ctype
open Rdigest
open Path

  (* strategy
     
     * Convert decls to (path/digest) source_declaration

       + They may contain some datatypes defined already.
         These dependent datatypes are
         - with digests (already calculated)
         - with manifest (path/digest) rtyp 
             (should be in "normal form" at this point) 
         - with flag of lazily calculated

       + Replace the dependent datatype paths with these infos
 
     * For each type decls, calculate type_digest info

       + Retrieve recursive dependency graph and 
         dependent nonrecursive types whose digest are not available.
  *)

let calc env assoc_list =
  Rdigest.debug ("Tdigest: " ^ catenate_sep " " (List.map (fun (id, _) -> Ident.name id) assoc_list));

  let id_list = List.map (fun (id, _) -> id) assoc_list in

  let get_rtype typ =
    let varcnt = ref 0 
    and vartbl = ref [] in
  	  
    let rec sub t = 
      let new_type_var t =
  	let rt = RTvar !varcnt in
  	incr varcnt;
  	vartbl := (t, rt) :: !vartbl;
  	rt
      in
  	      
      let t = full_expand env t in
      match t.desc with
  	Tvar -> 
  	  begin
  	    try List.assq t !vartbl with Not_found -> new_type_var t
  	  end
      | Tarrow (f,t) -> RTarrow (sub f, sub t) 
      | Ttuple tl -> RTtuple (List.map sub tl)
      | Tconstr (p,tl,_) -> 
  	  begin match p with
  	    Pident id when List.mem id id_list ->
              (* recursive *)
	      RTconstr (Path p, List.map sub tl)
  	  | _ ->
	      let tl' = List.map sub tl in
  	      begin match Env.find_type_digest p env with 
  		DDigest d ->
  		  RTconstr (Digest d, tl')
  	      | DManifest rtyp ->
  		  subst_rtype_at_var tl' (map_rtype (fun x -> RTvar x) (fun ci typs -> RTconstr (Digest ci, typs)) rtyp)
  	      | DSource _ | DLazy -> 
  		  RTconstr (Path p, tl')
	      |	DNone -> raise No_digest_info
  	      end
  	  end
      | Tlink t -> sub t
      | Tobject _ | Tfield _ | Tnil | Tany -> 
  	  raise (Dtype.Generic_unsupported_error "No objects, please")
      | Tdvar _ -> 
  	  raise (Dtype.Generic_system_error "Tdigest.get_rtype : Tdvar")
    in
    sub typ
  in

  let get_sdecl decl =
    (* check type_params *)
    List.iter (fun par -> 
      let par = repr par in
      if par.desc <> Tvar then begin
  	Printtyp.type_expr par;
  	Format.print_newline ();
  	raise (Dtype.Generic_system_error "Tdigest.calc : non tvar par!")
      end)
      decl.type_params;
      
    try
      match decl.type_kind with
	Type_abstract ->
  	  begin match decl.type_manifest with
  	    Some typ -> SType_manifest (get_rtype typ)
  	  | None -> SType_abstract
  	  end
      | Type_variant constructors ->
  	  SType_variant (List.map (fun (_, typs) -> List.map get_rtype typs)
  			   constructors )
      | Type_record members ->
  	  SType_record (List.map (fun (_, _, typ) -> get_rtype typ) members)
    with
      No_digest_info -> SType_none
  in

  let assoc_list = 
    List.map (fun (id, decl) -> id, (decl, get_sdecl decl)) assoc_list
  in
  
  (* create dependency informations *)

  let used_datatypes id sdecl = 

    let recursive = ref [] in
    let regular = ref [] in
    let add path buffer =
      if List.mem path !buffer then ()
      else buffer := path :: !buffer
    in

    let rec extract_rtype rtyp = 
      let rec sub = function
	  RTvar _ -> () 
	| RTarrow (f,t) -> sub f; sub t
	| RTtuple typs -> List.iter sub typs
	| RTconstr (pd, args) ->
	    List.iter sub args;
	    begin 
	      match pd with
		Path (Pident id) when List.mem id id_list ->
		  (* recursive *)
		  extract_sdecl id
	      | Path path -> add path regular
	      | Digest _ -> () 
	    end
      in
      sub rtyp

    and extract_sdecl id =
      let (_, sdecl) = List.assoc id assoc_list in
      if not (List.mem id !recursive) then begin
	add id recursive;
	match sdecl with
	  SType_abstract -> ()
	| SType_digest _ -> (* impossible *)
	    raise (Dtype.Generic_system_error "extract_sdecl SType_digest")
	| SType_manifest rtyp -> extract_rtype rtyp
	| SType_variant constructors ->
	    List.iter (List.iter extract_rtype) constructors
	| SType_record typs -> List.iter extract_rtype typs
	| SType_none -> raise No_digest_info
      end
    in
    (* find dependencies *)
    try
      extract_sdecl id;
      sdecl, !recursive, !regular
    with
      No_digest_info -> SType_none, [], []
  in

  let assoc_list = 
    List.map (fun (id, (decl, sdecl)) ->
      let sdecl, recursive, regular = used_datatypes id sdecl in
      (id, (decl, sdecl, recursive, regular))) assoc_list in

  let dependent_datatypes (id, (decl, sdecl, recursive, regular)) =
    let real_recursive = ref []
    and separatable = ref []
    in
    List.iter (fun id' ->
      let _, _, rec', _ = List.assoc id' assoc_list in
      if List.mem id rec' then 
	(* mutually recursive *)
	real_recursive := !real_recursive @ [id']
      else
	(* non-recursive, separatable actually *)
	separatable := !separatable @ [id']) recursive;
    id, (decl, sdecl, !real_recursive, !separatable, regular)
  in

  let assoc_list = List.map dependent_datatypes assoc_list in

  (* Sort them *)
  let final_list = 
    let rec get_level id = (* inefficient... *)
      let max = ref 0 in
      let (_, _, _, sep, _) = List.assoc id assoc_list in
      List.iter (fun id -> 
	let lev = get_level id in
	if lev > !max then max := lev ) sep;
      !max + 1
    in
    Sort.list (fun (id, _) (id', _) ->
      get_level id < get_level id') assoc_list
  in
  
  List.map (fun (id, (decl, sdecl, recursive, separatable, regular)) ->
    let type_digest = 
      match sdecl with
	SType_abstract -> 
Rdigest.debug (Printf.sprintf "\t%s = LAZY" (Ident.name id));
	  DLazy
      |	SType_none -> 
Rdigest.debug (Printf.sprintf "\t%s = NONE" (Ident.name id));
	  DNone
      | SType_digest _ -> (* impossible *)
Rdigest.debug (Printf.sprintf "\t%s Failure calc SType_digest" (Ident.name id));
	  raise (Failure "calc SType_digest")
      | SType_manifest rtyp -> 
	  begin
	    try
	      let drtyp = 
		map_rtype (fun x -> RTvar x) (fun ci typs ->
		  match ci with
		    Path _ -> raise Exit
		  | Digest d -> RTconstr (d, typs)) rtyp
	      in
Rdigest.debug (Printf.sprintf "\t%s = %s" (Ident.name id) (Rdigest.string_of_rtype (fun x -> x) drtyp));
	      DManifest drtyp
	    with
	      Exit ->
Rdigest.debug (Printf.sprintf "\t%s = DSource" (Ident.name id));
		let argpaths = ref [] in
  		iter_rtype (fun x -> ()) (fun ci typs ->
  		  match ci with 
  		    Path p -> argpaths := !argpaths @ [p]
		  | Digest d -> ()) rtyp;
		DSource { regular= !argpaths;
			  recursive= [];
			  declarations= [SType_manifest rtyp] }
	  end
      | SType_variant _ | SType_record _ ->
	  let string_of_source source =
	    let lazyflag = ref false in
	    let catenate_sep sep =
	      function 
		  [] -> ""
		| x::l -> List.fold_left (fun s s' -> s^sep^s') x l
	    in
	    let printer = function
		Path (Pident i) when List.mem i recursive -> 
		  "#" ^ string_of_int (Rdigest.list_mem_nth i recursive)
	      |	Path (Pident i) when List.mem i separatable -> 
		  begin
		    let (decl',_,_,_,_) = List.assoc i final_list in
		    match decl'.type_digest with
		      DDigest d -> d
		    | DSource _ -> 
			lazyflag := true;
			"[" ^ Ident.name i ^ "]"
		    | DNone -> raise (Dtype.Generic_system_error (Printf.sprintf "calc separatable digest %s is DNone, but ?" (Ident.name i)))
		    | _ -> raise (Dtype.Generic_system_error (Printf.sprintf "calc separatable digest for %s is not DDigest" (Ident.name i)))
		  end
	      |	Path p -> 
		  lazyflag := true;
		  "<" ^ Path.string_of p ^ ">"
	      | Digest d -> d
	    in
	    let string_of_sdecl = Rdigest.string_of_sdecl printer in
	    catenate_sep " and " (List.map string_of_sdecl source), !lazyflag
	  in
	  let source_string, lazyflag = 
	    string_of_source (List.map (fun id ->
	      let _, sdecl, _,_,_ = List.assoc id final_list in sdecl) 
				recursive)
	  in
	  if lazyflag then begin
	    DSource { regular= regular @ 
		        (List.map (fun id -> Pident id) separatable);
		      recursive= recursive;
		      declarations= List.map (fun id -> 
			let _, sdecl, _,_,_ = List.assoc id assoc_list 
			in sdecl) recursive }
	  end else begin
            (* No dependencies ! We can calculate it now ! *)
	    Rdigest.debug (Printf.sprintf "\t%s = %s" (Ident.name id) (Digest.string source_string));
	    DDigest ( Digest.string source_string )
	  end
    in
    decl.type_digest <- type_digest;
    (id, decl)) final_list

