open Misc
open Longident
open Path
open Parsetree
open Types
open Typedtree
open Asttypes

open Format

open Dtype

let debug = try ignore (Sys.getenv "DEBUG_GENERICTYPING"); true with _ -> false
let debug2 = try ignore (Sys.getenv "DEBUG_GENERICLOOP"); true with _ -> false

type patvar = Ident.t * Types.type_expr

type geninfo =
  | GenLet of Env.t * patvar list * patvar list * expression * geninfo list
  | GenGeneric of Env.t * patvar list * patvar * (generic_case * geninfo list) list
  | GenInst of expression  
  | GenEval of expression * type_expr * generic_konstraint list ref * geninfo list
;;

let pattern_variables pattern =
  let vars = ref [] in
  let rec sub pattern =
    match pattern.pat_desc with
    | Tpat_var id -> vars := (id,pattern.pat_type) :: !vars
    | Tpat_alias (p, id) -> sub p; vars := (id,pattern.pat_type) :: !vars
    | Tpat_tuple pats -> List.iter sub pats
    | Tpat_construct (_,pats) -> List.iter sub pats
    | Tpat_record lpats -> List.iter (fun (_,p) -> sub p) lpats
    | Tpat_array pats -> List.iter sub pats
    | Tpat_or (p1,p2) -> sub p1; sub p2
    | _ -> assert false
  in
  sub pattern;
  !vars
;;

let find_generics env strs =
  let rec find e =
    match e.exp_desc with
    | Texp_let (recflag,pexps,exp) ->
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,e) -> pattern_variables p) pexps)
	  else []
	in
	List.flatten (List.map (fun (p,e) -> 
	  if Dtype.generalized_dynamic_variables p.pat_type <> [] 
	  then begin
	    [GenLet(exp.exp_env, recursives, pattern_variables p, e, find e)]
	  end else find e) pexps) @
	find exp
    | Texp_generic (recflag, bindings, exp) ->
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,_) -> pattern_variables p) bindings)
	  else []
	in
	(List.map (fun (p,caseset) ->
	  let patvar = 
	    match pattern_variables p with
	    | [patvar] -> patvar
	    | _ -> assert false
	  in
	  GenGeneric (exp.exp_env,
		      recursives,
		      patvar,
		      List.map (fun cas -> 
			cas, find cas.gencase_method) 
			caseset.gen_cases)) bindings) @
	find exp
    | Texp_ident (path,vdesc,_) ->
	if Dtype.generalized_dynamic_variables vdesc.val_type <> []
	then [GenInst e] else []
    | Texp_constant _ -> []
    | Texp_function pexps ->
	List.flatten (List.map (fun (_,e) -> find e) pexps)
    | Texp_apply (e,exps) ->
	List.flatten (List.map find (e::exps))
    | Texp_match (e, pexps) ->
	find e @ (List.flatten (List.map (fun (_,e) -> find e) pexps))
    | Texp_try (e, pexps) ->
	find e @ (List.flatten (List.map (fun (_,e) -> find e) pexps))
    | Texp_tuple exps -> List.flatten (List.map find exps)
    | Texp_construct (_, exps) -> List.flatten (List.map find exps)
    | Texp_record (lexps, eopt) ->
      	List.flatten (List.map (fun (_,e) -> find e) lexps) @ 
	begin match eopt with
	  Some e -> find e
	| None -> []
	end
    | Texp_field (e,_) -> find e
    | Texp_setfield (e1, _, e2) -> find e1 @ find e2
    | Texp_array exps -> List.flatten (List.map find exps)
    | Texp_ifthenelse (e1,e2,eopt) ->
      	find e1 @ find e2 @
	begin match eopt with
	  Some e -> find e
	| None -> []
	end
    | Texp_sequence (e1,e2) -> find e1 @ find e2
    | Texp_while (e1, e2) -> find e1 @ find e2
    | Texp_for (_, e1, e2, _, e3) ->
	find e1 @ find e2 @ find e3
    | Texp_when (e1,e2) -> find e1 @ find e2
    | Texp_override (_, pexps) ->
      	List.flatten (List.map (fun (_,e) -> find e) pexps)
    | Texp_letmodule (_, (* module_expr is done *) _, e) -> find e
    | Texp_coerce (e, ptexps) ->
	find e @ (List.flatten (List.map (fun (_,_,e) -> find e) ptexps))
    | _ -> assert false
(*
    | Texp_send (e,_) -> find e
    | Texp_new _ -> []
    | Texp_instvar _ -> []
    | Texp_setinstvar (_,_,e) -> find e
*)
  in
  let find_str str =
    match str with
    | Tstr_eval (e,scm,kref) -> 
	if Dtype.generalized_dynamic_variables scm <> [] then
	  [GenEval (e, scm, kref, find e)]
	else find e
    | Tstr_value (recflag, pexps) -> 
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,_) -> pattern_variables p) pexps)
	  else []
	in
	List.flatten (List.map (fun (p,e) ->
	  if Dtype.generalized_dynamic_variables p.pat_type <> [] 
	  then [GenLet(env, recursives, pattern_variables p, e, find e)]
	  else find e) pexps)
    | Tstr_generic (_,recflag,bindings) ->
	let recursives =
	  if recflag = Recursive then 
	    List.flatten (List.map (fun (p,_) -> pattern_variables p) bindings)
	  else []
	in
	List.map (fun (p,caseset) ->
	  let patvar =
	    match pattern_variables p with
	    | [patvar] -> patvar
	    | _ -> assert false
	  in
	  GenGeneric (env, recursives, patvar,
		      (List.map (fun cas -> 
			cas, find cas.gencase_method) 
			 caseset.gen_cases))) bindings
    | Tstr_genextend (_,_,_,cases) -> assert false (* not yet *)
 (* | Tstr_module is done of its typing ? *)
    | _ -> [] (* assert false *) (* I do not think about classes... *)
  in
  List.flatten (List.map find_str strs)
;;

let get_ident exp =
  match exp.exp_desc with
  | Texp_ident (p, vdesc,_) -> p, vdesc
  | _ -> assert false
;;

let path_of gid = fst (get_ident gid)

let filter_pinsts gendvars pinsts = 

  let rec contents dvars =
    try
      List.iter (fun gendvar ->
	match gendvar.desc with
	| Tdvar (info,_) ->
	    List.iter (fun dvar ->
	      match dvar.desc with
	      | Tdvar (info',_) ->
		  if info == info' then raise Exit
	      | _ -> assert false) dvars
	| _ -> assert false) gendvars;
      false
    with
    | Exit -> true
  in

  List.fold_right (fun pinst (related,notrelated) ->
    let p, vdesc = get_ident pinst in
    let dv1 = Dtype.dynamic_variables pinst.exp_type in
    let dv2 = Dtype.dynamic_variables vdesc.val_type in
    if contents dv1 || contents dv2 then
      (pinst :: related, notrelated)
    else
      (related, pinst :: notrelated)) pinsts ([],[])

;;
      

let fix_derived konst =
  (* polyinsts dvars are different from their corresponding generalized
     dvars. they must be same in the konstraints *)
  let gendvars = Dtype.generalized_dynamic_variables konst.konst_case in
  let info_dvars = List.map (fun dvar -> 
    match dvar.desc with
    | Tdvar (x,_) -> x, dvar
    | _ -> assert false) gendvars 
  in
  let rec sub ty = 
    match ty.desc with
    | Tdvar (x,_) ->
	begin try
	  List.assq x info_dvars 
	with
	  Not_found -> ty
	end
    | Tarrow (t1,t2) -> { ty with desc = Tarrow( sub t1, sub t2 ) }
    | Ttuple tys -> { ty with desc = Ttuple( List.map sub tys ) }
    | Tconstr (p, tys, abref) ->
	 { ty with desc = Tconstr (p, List.map sub tys, abref) }
    | Tlink ty -> sub ty
    | Tvar -> ty
    | _ -> assert false
  in
  { konst with 
      konst_derived = List.map (fun (p,t) -> p, sub t) konst.konst_derived}
;;

exception Not_admissible of expression * type_expr

let debug_assign id k = 
  Printtyp.ident id;
  print_string " ";
  Printtyp.konstraints k;
  force_newline ()

let rec assign ginfos = 
  let rec sub = function
    | GenLet (env, recursives, patvars, exp, ginfos) ->

	(* global filtering *)
	let gendvars = 
	  List.flatten (List.map (fun (id,pat_type) ->
	    Dtype.generalized_dynamic_variables pat_type) patvars)
	in
	let related_all,notrelated = filter_pinsts gendvars (assign ginfos) in

	(* for each pattern variable *)

	List.iter (fun (id,pat_type) ->
	  let gendvars_id = Dtype.generalized_dynamic_variables pat_type in
	  let related_to_id,_ = filter_pinsts gendvars_id related_all in
	  
	  (* assign Inst_poly_generic id *)
	  let pos = ref 0 in
	  List.iter (fun exp -> 
	    match exp.exp_desc with
	    | Texp_ident (_,_,iref) ->
		begin match !iref with
		| Inst_normal -> 
		    iref := Inst_poly_generic (List.hd gendvars_id, !pos)
		| Inst_poly_generic _ ->
		    prerr_endline "A poly inst depends on more than one patvars!";
		    assert false
		| Inst_full_generic _ -> assert false
		end;
		incr pos
	    | _ -> assert false ) related_to_id;

	  let derived = List.map (fun e ->
	    let path, vdesc = get_ident e in
	    begin try 
	      ignore (Env.find_value path env)
	    with
	    | Not_found ->
		failwith (Path.string_of path ^ " WILL BE LOST");
	    end;
	    path, e.exp_type) related_to_id
	  in

	  let sizes = List.fold_right (fun e st ->
	    let path, vdesc = get_ident e in
	    match path with
	    | Pident id -> 
	      	begin try 
		  ignore (List.assoc id recursives);
	  	  (e, Dtype.size_of_instance_predicate 
		     env e.exp_type vdesc.val_type)
		  :: st
	      	with
	      	| Not_found -> st
	      	end
	    | _ -> st) related_to_id []
	  in

	  if List.mem (id,pat_type) recursives then begin
	    let size = Dtype.size_of_instance_predicate env pat_type pat_type 
	    in
	    List.iter (fun (e,size') ->
	      if not (Dtype.leq size' size) then 
		raise (Not_admissible (e,pat_type) ))
	      sizes;
	  end;

	  let vdesc = Env.find_value (Pident id) env in
	  vdesc.val_konstraints <-
	     [ fix_derived
		 { konst_case= pat_type;
		   konst_derived= derived } ];
	  if debug then debug_assign id vdesc.val_konstraints;

	  ) patvars;

	notrelated

    | GenEval ({exp_desc= Texp_ident (p,vdesc,_)}, scm, kref, ginfos) ->
	(* for queries of generic konstraints *)
	kref := vdesc.val_konstraints;
	if debug then debug_assign (Ident.create "top") !kref;
	[]

    | GenEval (exp, scm, kref, ginfos) ->
	let gendvars = Dtype.generalized_dynamic_variables scm in
	let related,notrelated = filter_pinsts gendvars (assign ginfos) in
	let derived = List.map (fun e ->
	  let path,vdesc = get_ident e in
	  path, e.exp_type) related
	in

	kref := 
	   [ fix_derived
	       { konst_case= scm;
		 konst_derived= derived } ];
	
	if debug then debug_assign (Ident.create "top") !kref;

	notrelated

    | GenGeneric (env, recursives, (id,pat_type), cases) ->
	let notrelated_list = ref [] in
	let konstraint =
	  List.map (fun (cas, ginfos) ->
	    let gendvars = Dtype.generalized_dynamic_variables 
		cas.gencase_scheme
	    in
	    let related,notrelated = filter_pinsts gendvars (assign ginfos) in

	  (* assign Inst_poly_generic id *)
	  let pos = ref 0 in
	  List.iter (fun exp -> 
	    match exp.exp_desc with
	    | Texp_ident (_,_,iref) ->
		begin match !iref with
		| Inst_normal -> 
		    iref := Inst_poly_generic (List.hd gendvars, !pos)
		| Inst_poly_generic _ ->
		    prerr_endline "A poly inst depends on more than one patvars!";
		    assert false
		| Inst_full_generic _ -> assert false
		end;
		incr pos
	    | _ -> assert false ) related;

	    notrelated_list := notrelated @ !notrelated_list;
	    let derived = List.map (fun e ->
	      let path,vdesc = get_ident e in
	      path, e.exp_type) related
	    in

	    let sizes = List.fold_right (fun e st ->
	      let path, vdesc = get_ident e in
	      match path with
	      | Pident id -> 
		  begin try 
		    ignore (List.assoc id recursives);
	  	    (e, Dtype.size_of_instance_predicate 
		      env e.exp_type vdesc.val_type)
		    :: st
		  with
		  |	Not_found -> st
		  end
	      | _ -> st) related []
	    in

	    if List.mem (id,pat_type) recursives then begin
	      let size = 
		Dtype.size_of_instance_predicate env cas.gencase_scheme pat_type 
	      in
	      List.iter (fun (e,size') ->
	      	if not (Dtype.leq size' size) then 
		  raise (Not_admissible (e,cas.gencase_scheme) ))
	      	sizes;
	    end;

	    fix_derived
	      { konst_case = cas.gencase_scheme;
		konst_derived= derived }
	      ) cases
	in

	let vdesc = Env.find_value (Pident id) env in
	vdesc.val_konstraints <- konstraint;
	if debug then debug_assign id vdesc.val_konstraints;
	!notrelated_list

    | GenInst exp -> [exp]
  in
  List.flatten (List.map sub ginfos)
;;
    
  
let assign ginfos =
  List.map (fun e -> e, ref (-1)) (assign ginfos)
;;

type type_infos = Type of type_expr | Konstraints of generic_konstraint list

let copy_konstraint cache konst =
  let cas = copy_type cache konst.konst_case in
  let derived = 
    List.map (fun (id,ty) -> (id, copy_type cache ty)) konst.konst_derived
  in
  { konst_case= cas;
    konst_derived= derived
  } 
;;

(******************************* Unification test (not actually unify types) *)

let unifiable env typ1 typ2 =
  (* Not really unify them, just look they are unifiable or not *)
  match copy_types Reset [typ1; typ2] with (* tdvar links are reset *)
  | [typ1'; typ2'] ->
      begin try Ctype.unify env typ1' typ2'; true with 
	Ctype.Unify _ -> false end
  | _ -> assert false
;;

exception No_match
exception No_match_with of Path.t * value_description * type_expr
exception Invalid_full_instance of 
  expression * Path.t * value_description * type_expr
exception Fixed of type_expr * generic_konstraint * int
exception Polyrec_restriction

let new_par () =
  force_newline ();
  print_string "  ";
  open_box 0
;;

let end_par () =
  close_box ();
  force_newline ();
;;

type log = { paths: Path.t list; 
	     type_grown: Path.t list;
	     cache: ((Path.t * type_expr) * int) list }

let empty_log = { paths= []; type_grown= []; cache= [] }

let rec print_flow = function
  | Flow_node node ->
      printf "#%d(id=%d) " node.flow_pos node.flow_id;
      open_box 0;
      List.iter (fun f -> print_flow f; force_newline ()) node.flow_sub;
      close_box ()
  | Flow_loop id -> printf "loop=>%d" id
  | Flow_type typs -> 
      let first = ref true in
      print_string "<<"; 
      List.iter (fun typ -> 
	if not !first then print_string ", ";
	first := true;
	Printtyp.type_expr typ) typs; 
      print_string ">>"
;;

let node_id = ref 0

let mode_choose_first_match = ref false
let mode_old = false

(* Check t <: \sigma d *)
let rec reconst_konst_sub log env path typ konst =
  if debug then begin
    print_string "Check  ";
    Printtyp.path path; 
    print_string " : ";
    Printtyp.type_expr typ;
    print_string " <: ";
    Printtyp.konstraints [konst];
    new_par ();
  end;
  
  let copy_cache = create_copy_cache Duplicate in
  
  let orgtyp = typ in
  let orgkonst = konst in

  let typ = copy_type copy_cache typ in
  let log = { log with cache = List.map (fun ((p,t),id) -> 
                          (p, copy_type copy_cache t),id) log.cache} in
  let konst = copy_konstraint copy_cache konst in

  let is_instance =
    Ctype.moregeneral env true konst.konst_case typ
  in

  let dvars = dynamic_variables typ in

  try
    begin try
      Ctype.unify env typ konst.konst_case; (* may fail *)
    with e ->
      if debug then print_string "Not unifiable";
      raise e
    end;

    let type_grown =
      try List.iter (fun dv ->
	let s = Dtype.size dv in 
	let grown = 
	  match List.length s.tdvars with
	  | 0 -> false
	  | 1 -> s.const <> 0
	  | _ -> true
	in
	if grown then raise Exit) dvars;
  	false
      with
      | Exit -> true
    in

    let log =
      if type_grown then begin
	if debug then begin
	  print_string "TYPE SIZE GROWN";
	  force_newline ();
	end;
        if List.mem path log.type_grown then begin
	  if debug then begin
	    print_string "POLYREC RESTRICTION !!!";
	    force_newline ();
	  end;
	  raise Polyrec_restriction
	end else {log with type_grown= path :: log.type_grown}
      end else log
    in  

    if debug then begin
      if is_instance then print_string "Instance  "
      else print_string "Unifiable  ";
      Printtyp.type_expr orgtyp;
      if is_instance then print_string " of  "
      else print_string " with ";
      Printtyp.type_scheme orgkonst.konst_case;
      if orgkonst.konst_derived <> [] then begin
	force_newline ();
	print_string "Derived konstraint check";
	force_newline ();
      end;
    end;

    let amb, flows_rev = List.fold_left (fun (amb,frev) (path,t) -> 
      let vdesc = 
	try
	  Env.find_value path env 
	with
	| Not_found ->
	    print_string "!!!";
	    Printtyp.path path;
	    print_string " is not found in env";
	    force_newline ();
	    raise Not_found
      in 
      let _,a,flow = reconst_path log env t path vdesc in (* may fail *) 
      amb + a, flow :: frev) (0,[]) konst.konst_derived
    in
 
    let flows = 
      try
	Some (List.rev (List.map (function 
	  | Some f -> f 
	  | None -> raise Exit) flows_rev))
      with
      |	Exit -> None
    in

    let amb = amb + 
              if is_instance || 
              (!mode_choose_first_match (* fake as if it is an instance *)
		 && flows <> None) (* only when the flow is clear *)
              then 0 else 1 
    in
    

    if debug then begin
      end_par ();

      printf "%d: " amb;
      Printtyp.type_expr typ;
      print_string " <: ";
      Printtyp.konstraints [konst];
      force_newline ();
      force_newline ();
    end;

    amb, typ, flows
  with
  | e -> 
      if debug then end_par ();
      if mode_old && is_instance then raise Exit;
      if debug then begin
	printf "X: ";
	Printtyp.type_expr orgtyp;
	print_string " <: ";
	Printtyp.konstraints [orgkonst];
	force_newline ();
	force_newline ();
      end;
      raise e

(* Check t <: [..|..|..] *)
and reconst_konst log env path typ konsts =

  (* cached ? *)
  let copy_cache = create_copy_cache Duplicate in 

  let orgtyp = typ in
  let orgkonsts = konsts in

  let typ = copy_type copy_cache typ in
  let log = { log with cache= List.map (fun ((p,t),id) ->
                  (p, copy_type copy_cache t),id) log.cache } in
  let konsts = List.map (copy_konstraint copy_cache) konsts in

  let found = ref [] in 

  let amb = 
    try
      let pos = ref (-1) in
      List.fold_left (fun amb k ->
	incr pos;
  	try
  	  let a,t,flows = reconst_konst_sub log env path typ k in
  	  found := (!pos,a,t,flows) :: !found;
  	  if amb = 0 && a = 0 then begin
	    if debug then begin
	      print_string "THIS CASE MUST BE SELECTED";
	      force_newline ();
	    end;
	    raise Exit; (* The first match is fixed *)
	  end;
  	  amb + a 
  	with
  	| e when e <> Exit -> amb) 0 konsts
    with
      Exit -> 0
  in
  
  let anti, total_amb, flow =
    match !found  with
    | [] -> 
	if debug then begin
	  print_string "no match";
	  force_newline ();
	end;
	raise No_match
    | [pos,a,t,flows] -> 
	if debug then begin
	  print_string "1 match: ";
	  printf "%d: " a;
	  Printtyp.type_expr t;
	  force_newline ();
	end;
	t, a, 
	begin match flows with
	| None -> None
	| Some flows -> Some (Flow_node {flow_pos= pos; flow_sub= flows; flow_id= -1})
	end
    | _ -> 
	let atype =
	  Dtype.anti_unify_types env 
	    (List.map (fun (_,_,t,_) -> t) !found)
	in

	if debug then begin
	  printf "%d matches: " (List.length !found);
	  printf "%d: " (amb + List.length !found);
	  Printtyp.type_expr atype;
	  force_newline ();
	end;

	atype, amb + List.length !found, None
  in

  if debug then begin
    print_string "unify: ";
    Printtyp.type_expr typ;
    print_string " ~= ";
    Printtyp.type_expr anti;
  end;
    Ctype.unify env typ anti;

  if debug then begin
    print_string " ==> ";
    Printtyp.type_expr typ;
    force_newline ();
  end;

  total_amb, typ, flow

(* Check p : t <: \sigma K *)
and reconst_path log env typ path vdesc =

  if debug then begin
    print_string "Check  ";
    Printtyp.path path;
    print_string " : ";
    Printtyp.type_expr typ;
    print_string " <: ";
    Printtyp.konstraints vdesc.val_konstraints;
  
    new_par ();
  end;

  if vdesc.val_konstraints = [] then begin
    if debug then begin
      end_par ();
      print_string "This is generic primitive";
      force_newline ();
    end;

    let copy_cache = create_copy_cache Duplicate in
    let typ = copy_type copy_cache typ in
    let vtype = copy_type copy_cache vdesc.val_type in
    let dvars = generalized_dynamic_variables vtype in
    begin try
      if debug then begin
	Printtyp.type_scheme vtype;
	Format.print_string " ~=~ ";
	Printtyp.type_scheme typ;
	Format.print_newline ();
      end;
      Ctype.unify env typ vtype
    with
      _ -> failwith "This unification must suceed!"
    end;
    false, 0, Some (Flow_type dvars)
  end else begin

    let my_id = !node_id in
    incr node_id;
    try
      let amb,t,flow = 
	let rec assoc_cache (path,typ) = function
	  | ((p,t),id) :: xs ->
	      if path = p then begin
(*
  print_string "find ";
  Printtyp.path path;
  print_string ",";
  Printtyp.type_expr typ;
  print_string " = ";
  Printtyp.path p;
  print_string ",";
  Printtyp.type_expr t;
  force_newline ();
*)
		if Ctype.equal env false [typ] [t] then id
		else assoc_cache (path,typ) xs
	      end else assoc_cache (path,typ) xs
	  | [] -> raise Not_found
	in
	try
	  let id = assoc_cache (path, typ) log.cache in
	  if debug then begin
	    print_string "FOUND IN THE CACHE";
	    force_newline ();
	  end;
	  0,typ, Some (Flow_loop id)
	with
	| Not_found ->
	    let log = 
	      let paths, rec_check =
	      	if List.mem path log.paths then log.paths, true
  	      	else path :: log.paths, false
	      in
	      { paths= paths;
	      	type_grown= log.type_grown;
	      	cache= ((path,typ),my_id)::log.cache }
	    in
	    reconst_konst log env path typ vdesc.val_konstraints 
      in
  
      begin match flow with
      |	Some (Flow_node node) -> node.flow_id <- my_id
      |	Some (Flow_loop _) -> ()
      |	None -> ()
      |	_ -> assert false
      end;

      if debug then end_par ();

      (* check that typ will be changed... *)
      (* possible bug: t may contain free variables occur in the program *)

      let typ' = Dtype.force_generalize_copy typ 
      and t' = Dtype.force_generalize_copy t in
      let type_changed = not (Ctype.moregeneral env false t' typ') in

      Ctype.unify env t typ;

      if debug then begin
	printf "%d: " amb;
	Printtyp.path path;
	print_string " : ";
	Printtyp.type_expr typ;
(*
        print_string " <: ";
        Printtyp.extended_scheme vdesc.val_type vdesc.val_konstraints;
*)
	force_newline ();
      end;

      type_changed, amb, flow
    with
    | No_match -> 
  	if debug then end_par ();
  	raise (No_match_with (path, vdesc, typ))
  end
;;

let reconst_exp (e,res) =
  let path, vdesc = get_ident e in
  try
    let changed, amb, flow = reconst_path empty_log e.exp_env  e.exp_type path vdesc in
    res := amb;
    changed, amb, flow
  with
    No_match_with (p,v,t) ->
      raise (Invalid_full_instance (e,p,v,t))
  
let reconst exps = 
  let changed, amb, flows_rev = 
    List.fold_left (fun (ch,amb,frev) e -> 
      let changed, a, flow = reconst_exp e in
      ch || changed, a + amb, flow :: frev) (false,0,[]) exps
  in
  changed, amb, List.rev flows_rev 
;;

let reconst_loop choose_first_match exps =
  node_id := 0;
  mode_choose_first_match := choose_first_match;
  let prev_amb = ref None in
  let decreased amb = 
    if amb = 0 then false
    else
      match !prev_amb with
      | None -> 
  	  prev_amb := Some amb;
  	  true
      | Some p -> 
	  if amb > p then begin
	    prerr_endline "AMBIGUOUSITY RAISED";
	    true
	  end else begin
  	    let r = p > amb in
  	    prev_amb := Some amb;
  	    r
	  end
  in

  let rec loop () =
    if debug then new_par ();
    let changed, amb, flows = reconst exps in 

    if debug then begin
      end_par ();
      print_newline ();
    end;

    if debug2 then begin
      if exps <> [] then begin
  	List.iter (fun (exp,_) ->
  	  let path, vdesc = get_ident exp in
  	  Printtyp.path path;
  	  print_string " : ";
  	  Printtyp.type_expr exp.exp_type;
  	  print_newline () ) exps;
  	print_string "---";
  	print_newline ();
      end
    end;

    if changed (* decreased amb *) 
	&& not choose_first_match (* if we loop with choose_first_match,
				     it may loop forever.
				     Ex. curry (fun x -> x)
				     *)
    then begin
      if debug then begin
	printf "(%d) TRY AGAIN---------" amb;
	print_newline ();
      end;
      loop ()
    end else begin
      if debug then begin
	printf "(%d) FIXED POINT---------" amb;
	print_newline ();
      end;
      amb, flows
    end
  in
  loop ()
;;

let unify_free_dvar_to_dyn exps =
  let modified = ref false in
  List.iter (fun (e,res) ->
    let path, vdesc = get_ident e in

    let rec unifier ty =
      begin match ty.desc with
      | Tdvar (x,_) when !x = Not_generalized ->
	  Ctype.unify e.exp_env ty Predef.type_dyn;
	  modified := true
      | _ -> ()
      end;
      Btype.iter_type_expr unifier ty
    in
    unifier e.exp_type) exps;
  !modified
;;

let reconst_str env strs =
  let geninfo = find_generics env strs in

  if geninfo <> [] then begin
    if debug || debug2 then begin
      Format.print_string "GENERIC TYPING";
      Format.print_newline ();
  
      Format.print_string "ASSIGN GENINFO----------";
      Format.print_newline ();
    end;

    let fullinsts = assign geninfo in
  
    if fullinsts <> [] then begin
      let amb,flows =
  	if debug || debug2 then begin
	  Format.print_string "FIRST LOOP----------";
  	  Format.print_newline ();
	end;
  	let amb,flows = reconst_loop false fullinsts in

  	let amb,flows =
	  if amb = 0 then amb, flows
  	  else begin
  	    if debug || debug2 then begin
	      Format.print_string "SECOND LOOP (Guess further)----------";
  	      Format.print_newline ();
	    end;
  	    reconst_loop true fullinsts
	  end
	in

  	ignore (unify_free_dvar_to_dyn fullinsts);
  	if debug || debug2 then begin
	  Format.print_string "THIRD LOOP (Final check with DYN unification)----------";
  	  Format.print_newline ();
	end;
  	reconst_loop false fullinsts;
      in
    
      List.iter2 (fun (exp,_) flow ->
  	match exp.exp_desc with
  	| Texp_ident (_,_,flowref) -> 
  	    begin match flow with
  	    | Some flow -> flowref := Inst_full_generic flow
  	    | None -> assert false
  	    end
  	| _ -> assert false)  fullinsts flows;
    
    
      if debug then begin
	List.iter2 (fun (exp,_) flow ->
  	  let p, vdesc = get_ident exp in
  	  let flow =
  	    match flow with
  	    | Some flow -> flow
  	    | None -> assert false
  	  in
  	  Printtyp.path p;
  	  Format.print_string " : ";
  	  Printtyp.type_expr exp.exp_type;
  	  Format.print_string " --- ";
  	  Format.open_box 0;
  	  print_flow flow;
  	  Format.close_box ();
  	  Format.force_newline ()) fullinsts flows;
    
	Format.print_string "Generic typing finished";
	Format.print_newline ();
      end;

    end
  end
;;

let now_typing = ref false

let reconst_str env strs =
  Ctype.under_generic_typing := true;
  try
    let result = reconst_str env strs in
    Ctype.under_generic_typing := false;
    result
  with
    e ->
        Ctype.under_generic_typing := false;
      raise e
